import subprocess
import sys

import pytest

from mx_bluesky import __version__


@pytest.mark.skip_in_pycharm(reason="subprocess returns tty escape sequences")
def test_cli_version():
    cmd = [sys.executable, "-m", "mx_bluesky", "--version"]
    assert subprocess.check_output(cmd).decode().strip() == __version__



from __future__ import annotations

import numpy
from dodal.devices.detector import DetectorParams
from dodal.devices.oav import utils as oav_utils

from mx_bluesky.common.external_interaction.ispyb.data_model import (
    DataCollectionGridInfo,
    DataCollectionInfo,
)


def populate_xz_data_collection_info(detector_params: DetectorParams):
    assert (
        detector_params.omega_start is not None
        and detector_params.run_number is not None
    ), "StoreGridscanInIspyb failed to get parameters"
    run_number = detector_params.run_number + 1
    info = DataCollectionInfo(
        data_collection_number=run_number,
    )
    return info


def populate_xy_data_collection_info(detector_params: DetectorParams):
    return DataCollectionInfo(
        data_collection_number=detector_params.run_number,
    )


def construct_comment_for_gridscan(grid_info: DataCollectionGridInfo) -> str:
    assert grid_info is not None, "StoreGridScanInIspyb failed to get parameters"

    bottom_right = oav_utils.bottom_right_from_top_left(
        numpy.array(
            [grid_info.snapshot_offset_x_pixel, grid_info.snapshot_offset_y_pixel]
        ),  # type: ignore
        grid_info.steps_x,
        grid_info.steps_y,
        grid_info.dx_in_mm,
        grid_info.dy_in_mm,
        grid_info.microns_per_pixel_x,
        grid_info.microns_per_pixel_y,
    )
    return (
        "Diffraction grid scan of "
        f"{grid_info.steps_x} by "
        f"{grid_info.steps_y} images in "
        f"{(grid_info.dx_in_mm * 1e3):.1f} um by "
        f"{(grid_info.dy_in_mm * 1e3):.1f} um steps. "
        f"Top left (px): [{int(grid_info.snapshot_offset_x_pixel)},{int(grid_info.snapshot_offset_y_pixel)}], "
        f"bottom right (px): [{bottom_right[0]},{bottom_right[1]}]."
    )



from unittest.mock import MagicMock, patch

import pytest
from bluesky.callbacks import CallbackBase
from bluesky.plan_stubs import null
from bluesky.run_engine import RunEngine
from bluesky.simulators import RunEngineSimulator, assert_message_and_return_remaining
from bluesky.utils import MsgGenerator
from dodal.beamlines.i03 import eiger
from dodal.devices.fast_grid_scan import ZebraFastGridScan
from dodal.devices.synchrotron import Synchrotron, SynchrotronMode
from dodal.devices.zocalo.zocalo_results import (
    ZOCALO_STAGE_GROUP,
)
from event_model.documents import Event, RunStart
from ophyd_async.core import init_devices
from ophyd_async.testing import set_mock_value

from mx_bluesky.common.parameters.constants import (
    PlanNameConstants,
)
from mx_bluesky.common.plans.do_fgs import kickoff_and_complete_gridscan


@pytest.fixture
def fgs_devices(RE):
    with init_devices(mock=True):
        synchrotron = Synchrotron()
        grid_scan_device = ZebraFastGridScan("zebra_fgs")

    # Eiger done separately as not ophyd-async yet
    detector = eiger(mock=True)

    return {
        "synchrotron": synchrotron,
        "grid_scan_device": grid_scan_device,
        "detector": detector,
    }


@patch("mx_bluesky.common.plans.do_fgs.read_hardware_for_zocalo")
@patch("mx_bluesky.common.plans.do_fgs.check_topup_and_wait_if_necessary")
def test_kickoff_and_complete_gridscan_correct_messages(
    mock_check_topup,
    mock_read_hardware,
    sim_run_engine: RunEngineSimulator,
    fgs_devices,
):
    def null_plan() -> MsgGenerator:
        yield from null()

    synchrotron = fgs_devices["synchrotron"]
    detector = fgs_devices["detector"]
    fgs_device = fgs_devices["grid_scan_device"]

    msgs = sim_run_engine.simulate_plan(
        kickoff_and_complete_gridscan(
            fgs_device,
            detector,
            synchrotron,
            scan_points=[],
            scan_start_indices=[],
            plan_during_collection=null_plan,
        )
    )

    msgs = assert_message_and_return_remaining(
        msgs,
        lambda msg: msg.command == "read"
        and msg.obj.name == "grid_scan_device-expected_images",
    )

    msgs = assert_message_and_return_remaining(
        msgs,
        lambda msg: msg.command == "read" and msg.obj.name == "eiger_cam_acquire_time",
    )

    mock_check_topup.assert_called_once()
    mock_read_hardware.assert_called_once()

    msgs = assert_message_and_return_remaining(msgs, lambda msg: msg.command == "wait")

    msgs = assert_message_and_return_remaining(
        msgs,
        lambda msg: msg.command == "wait" and msg.kwargs["group"] == ZOCALO_STAGE_GROUP,
    )

    msgs = assert_message_and_return_remaining(
        msgs, lambda msg: msg.command == "kickoff"
    )

    msgs = assert_message_and_return_remaining(msgs, lambda msg: msg.command == "wait")

    msgs = assert_message_and_return_remaining(msgs, lambda msg: msg.command == "null")

    msgs = assert_message_and_return_remaining(
        msgs,
        lambda msg: msg.command == "complete" and msg.obj.name == "grid_scan_device",
    )

    msgs = assert_message_and_return_remaining(msgs, lambda msg: msg.command == "wait")


# This test should use the real Zocalo callbacks once https://github.com/DiamondLightSource/mx-bluesky/issues/215 is done
def test_kickoff_and_complete_gridscan_with_run_engine_correct_documents(
    RE: RunEngine, fgs_devices
):
    class TestCallback(CallbackBase):
        def start(self, doc: RunStart):
            self.subplan_name = doc.get("subplan_name")
            self.scan_points = doc.get("scan_points")
            self.scan_start_indices = doc.get("scan_start_indices")

        def event(self, doc: Event):
            self.event_data = list(doc.get("data").keys())
            return doc

    test_callback = TestCallback()

    RE.subscribe(test_callback)
    synchrotron = fgs_devices["synchrotron"]
    set_mock_value(synchrotron.synchrotron_mode, SynchrotronMode.DEV)
    detector = fgs_devices["detector"]
    fgs_device: ZebraFastGridScan = fgs_devices["grid_scan_device"]

    detector.unstage = MagicMock()

    set_mock_value(fgs_device.status, 1)

    with patch("mx_bluesky.common.plans.do_fgs.bps.complete"):
        RE(
            kickoff_and_complete_gridscan(
                fgs_device,
                detector,
                synchrotron,
                scan_points=[],
                scan_start_indices=[],
            )
        )

    assert test_callback.subplan_name == PlanNameConstants.DO_FGS
    assert test_callback.scan_points == []
    assert test_callback.scan_start_indices == []
    assert len(test_callback.event_data) == 1
    assert test_callback.event_data[0] == "eiger_odin_file_writer_id"


@patch("mx_bluesky.common.plans.do_fgs.check_topup_and_wait_if_necessary")
def test_error_if_kickoff_and_complete_gridscan_parameters_wrong_lengths(
    mock_check_topup, sim_run_engine: RunEngineSimulator, fgs_devices
):
    synchrotron = fgs_devices["synchrotron"]
    detector = fgs_devices["detector"]
    fgs_device = fgs_devices["grid_scan_device"]
    with pytest.raises(AssertionError):
        sim_run_engine.simulate_plan(
            kickoff_and_complete_gridscan(
                fgs_device,
                detector,
                synchrotron,
                scan_points=[],
                scan_start_indices=[0],
            )
        )



import logging
from collections.abc import Callable, Sequence
from threading import Thread

import bluesky.plan_stubs as bps
from bluesky.callbacks import CallbackBase
from bluesky.callbacks.zmq import Proxy, RemoteDispatcher
from dodal.log import LOGGER as dodal_logger
from dodal.log import set_up_all_logging_handlers

from mx_bluesky.common.external_interaction.callbacks.common.log_uid_tag_callback import (
    LogUidTaggingCallback,
)
from mx_bluesky.common.external_interaction.callbacks.common.zocalo_callback import (
    ZocaloCallback,
)
from mx_bluesky.common.external_interaction.callbacks.sample_handling.sample_handling_callback import (
    SampleHandlingCallback,
)
from mx_bluesky.common.external_interaction.callbacks.xray_centre.ispyb_callback import (
    GridscanISPyBCallback,
)
from mx_bluesky.common.external_interaction.callbacks.xray_centre.nexus_callback import (
    GridscanNexusFileCallback,
)
from mx_bluesky.common.utils.log import (
    ISPYB_ZOCALO_CALLBACK_LOGGER,
    NEXUS_LOGGER,
    _get_logging_dir,
    tag_filter,
)
from mx_bluesky.hyperion.external_interaction.callbacks.robot_load.ispyb_callback import (
    RobotLoadISPyBCallback,
)
from mx_bluesky.hyperion.external_interaction.callbacks.rotation.ispyb_callback import (
    RotationISPyBCallback,
)
from mx_bluesky.hyperion.external_interaction.callbacks.rotation.nexus_callback import (
    RotationNexusFileCallback,
)
from mx_bluesky.hyperion.external_interaction.callbacks.snapshot_callback import (
    BeamDrawingCallback,
)
from mx_bluesky.hyperion.parameters.cli import parse_callback_dev_mode_arg
from mx_bluesky.hyperion.parameters.constants import CONST
from mx_bluesky.hyperion.parameters.gridscan import (
    GridCommonWithHyperionDetectorParams,
    HyperionSpecifiedThreeDGridScan,
)

LIVENESS_POLL_SECONDS = 1
ERROR_LOG_BUFFER_LINES = 5000


def create_gridscan_callbacks() -> tuple[
    GridscanNexusFileCallback, GridscanISPyBCallback
]:
    return (
        GridscanNexusFileCallback(param_type=HyperionSpecifiedThreeDGridScan),
        GridscanISPyBCallback(
            param_type=GridCommonWithHyperionDetectorParams,
            emit=ZocaloCallback(CONST.PLAN.DO_FGS, CONST.ZOCALO_ENV),
        ),
    )


def create_rotation_callbacks() -> tuple[
    RotationNexusFileCallback, RotationISPyBCallback
]:
    return (
        RotationNexusFileCallback(),
        RotationISPyBCallback(
            emit=ZocaloCallback(CONST.PLAN.ROTATION_MULTI, CONST.ZOCALO_ENV)
        ),
    )


def setup_callbacks() -> list[CallbackBase]:
    rot_nexus_cb, rot_ispyb_cb = create_rotation_callbacks()
    snapshot_cb = BeamDrawingCallback(emit=rot_ispyb_cb)
    return [
        *create_gridscan_callbacks(),
        rot_nexus_cb,
        snapshot_cb,
        LogUidTaggingCallback(),
        RobotLoadISPyBCallback(),
        SampleHandlingCallback(),
    ]


def setup_logging(dev_mode: bool):
    for logger, filename in [
        (ISPYB_ZOCALO_CALLBACK_LOGGER, "hyperion_ispyb_callback.log"),
        (NEXUS_LOGGER, "hyperion_nexus_callback.log"),
    ]:
        if logger.handlers == []:
            handlers = set_up_all_logging_handlers(
                logger,
                _get_logging_dir(),
                filename,
                dev_mode,
                error_log_buffer_lines=ERROR_LOG_BUFFER_LINES,
                graylog_port=CONST.GRAYLOG_PORT,
            )
            handlers["graylog_handler"].addFilter(tag_filter)
    log_info(f"Loggers initialised with dev_mode={dev_mode}")
    nexgen_logger = logging.getLogger("nexgen")
    nexgen_logger.parent = NEXUS_LOGGER
    dodal_logger.parent = ISPYB_ZOCALO_CALLBACK_LOGGER
    log_debug("nexgen logger added to nexus logger")


def setup_threads():
    proxy = Proxy(*CONST.CALLBACK_0MQ_PROXY_PORTS)
    dispatcher = RemoteDispatcher(f"localhost:{CONST.CALLBACK_0MQ_PROXY_PORTS[1]}")
    log_debug("Created proxy and dispatcher objects")

    def start_proxy():
        proxy.start()

    def start_dispatcher(callbacks: list[Callable]):
        [dispatcher.subscribe(cb) for cb in callbacks]
        dispatcher.start()

    return proxy, dispatcher, start_proxy, start_dispatcher


def log_info(msg, *args, **kwargs):
    ISPYB_ZOCALO_CALLBACK_LOGGER.info(msg, *args, **kwargs)
    NEXUS_LOGGER.info(msg, *args, **kwargs)


def log_debug(msg, *args, **kwargs):
    ISPYB_ZOCALO_CALLBACK_LOGGER.debug(msg, *args, **kwargs)
    NEXUS_LOGGER.debug(msg, *args, **kwargs)


def wait_for_threads_forever(threads: Sequence[Thread]):
    alive = [t.is_alive() for t in threads]
    try:
        log_debug("Trying to wait forever on callback and dispatcher threads")
        while all(alive):
            yield from bps.sleep(LIVENESS_POLL_SECONDS)
            alive = [t.is_alive() for t in threads]
    except KeyboardInterrupt:
        log_info("Main thread received interrupt - exiting.")
    else:
        log_info("Proxy or dispatcher thread ended - exiting.")


class HyperionCallbackRunner:
    """Runs Nexus, ISPyB and Zocalo callbacks in their own process."""

    def __init__(self, dev_mode) -> None:
        setup_logging(dev_mode)
        log_info("Hyperion callback process started.")

        self.callbacks = setup_callbacks()
        self.proxy, self.dispatcher, start_proxy, start_dispatcher = setup_threads()
        log_info("Created 0MQ proxy and local RemoteDispatcher.")

        self.proxy_thread = Thread(target=start_proxy, daemon=True)
        self.dispatcher_thread = Thread(
            target=start_dispatcher, args=[self.callbacks], daemon=True
        )

    def start(self):
        log_info(f"Launching threads, with callbacks: {self.callbacks}")
        self.proxy_thread.start()
        self.dispatcher_thread.start()
        log_info("Proxy and dispatcher thread launched.")
        wait_for_threads_forever([self.proxy_thread, self.dispatcher_thread])


def main(dev_mode=False) -> None:
    dev_mode = dev_mode or parse_callback_dev_mode_arg()
    print(f"In dev mode: {dev_mode}")
    runner = HyperionCallbackRunner(dev_mode)
    runner.start()


if __name__ == "__main__":
    main()



import asyncio
from typing import Literal
from unittest.mock import MagicMock, patch

import numpy as np
import pytest
from bluesky.run_engine import RunEngine
from dodal.beamlines import i03
from ophyd.status import Status
from ophyd_async.core import AsyncStatus
from ophyd_async.testing import set_mock_value

from mx_bluesky.common.utils.log import LOGGER
from mx_bluesky.hyperion.experiment_plans import optimise_attenuation_plan
from mx_bluesky.hyperion.experiment_plans.optimise_attenuation_plan import (
    AttenuationOptimisationFailedException,
    Direction,
    OptimizeAttenuationComposite,
    calculate_new_direction,
    check_parameters,
    deadtime_calc_new_transmission,
    deadtime_optimisation,
    is_counts_within_target,
    is_deadtime_optimised,
    total_counts_optimisation,
)


@pytest.fixture
def mock_emit():
    import logging

    test_handler = logging.Handler()
    test_handler.emit = MagicMock()  # type: ignore
    LOGGER.addHandler(test_handler)

    yield test_handler.emit

    LOGGER.removeHandler(test_handler)


@pytest.fixture
async def fake_composite(attenuator) -> OptimizeAttenuationComposite:
    sample_shutter = i03.sample_shutter(connect_immediately=True, mock=True)
    xspress3mini = i03.xspress3mini(connect_immediately=True, mock=True)

    return OptimizeAttenuationComposite(
        sample_shutter=sample_shutter, xspress3mini=xspress3mini, attenuator=attenuator
    )


def get_good_status():
    status = Status()
    status.set_finished()
    return status


@pytest.fixture
def fake_composite_mocked_sets(fake_composite: OptimizeAttenuationComposite):
    with (
        patch.object(
            fake_composite.xspress3mini,
            "stage",
            MagicMock(return_value=get_good_status()),
        ),
        patch.object(
            fake_composite.sample_shutter,
            "set",
            MagicMock(return_value=get_good_status()),
        ),
    ):
        yield fake_composite


def test_is_deadtime_optimised_returns_true_once_direction_is_flipped_and_deadtime_goes_back_above_threshold(
    RE: RunEngine,
):
    deadtime: float = 1
    direction = Direction.POSITIVE
    for _ in range(5):
        assert is_deadtime_optimised(deadtime, 0.5, 0.5, 1, Direction.POSITIVE) is False
        direction = calculate_new_direction(direction, deadtime, 0.5)
        deadtime -= 0.1
    assert direction == Direction.NEGATIVE
    deadtime = 0.4
    assert is_deadtime_optimised(deadtime, 0.5, 0.5, 1, direction) is True


def test_is_deadtime_is_optimised_logs_warning_when_upper_transmission_limit_is_reached(
    mock_emit: MagicMock,
):
    is_deadtime_optimised(0.5, 0.4, 0.9, 0.9, Direction.POSITIVE)
    latest_record = mock_emit.call_args.args[-1]
    assert latest_record.levelname == "WARNING"


def test_total_counts_calc_new_transmission_raises_warning_on_high_transmission(
    RE: RunEngine,
    mock_emit: MagicMock,
    fake_composite_mocked_sets: OptimizeAttenuationComposite,
):
    set_mock_value(
        fake_composite_mocked_sets.xspress3mini.dt_corrected_latest_mca[1],
        np.array([1, 1, 1, 1, 1, 1]),
    )
    RE(
        total_counts_optimisation(
            fake_composite_mocked_sets,
            transmission=0.1,
            low_roi=0,
            high_roi=1,
            lower_count_limit=0,
            upper_count_limit=0.1,
            target_count=1,
            max_cycles=1,
            upper_transmission_limit=0.1,
            lower_transmission_limit=0,
        )
    )

    latest_record = mock_emit.call_args.args[-1]
    assert latest_record.levelname == "WARNING"


@pytest.mark.parametrize(
    "old_direction, deadtime, deadtime_threshold, new_direction",
    [
        (Direction.POSITIVE, 0.1, 0.9, Direction.POSITIVE),
        (Direction.NEGATIVE, 0.5, 0.4, Direction.NEGATIVE),
    ],
)
def test_calculate_new_direction_gives_correct_value(
    old_direction: Direction | Direction,
    deadtime: float,
    deadtime_threshold: float,
    new_direction: Direction | Direction,
):
    assert (
        calculate_new_direction(old_direction, deadtime, deadtime_threshold)
        == new_direction
    )


@patch(
    "mx_bluesky.hyperion.experiment_plans.optimise_attenuation_plan.do_device_optimise_iteration",
    autospec=True,
)
def test_deadtime_optimisation_calculates_deadtime_correctly(
    mock_do_device_optimise_iteration,
    RE: RunEngine,
    fake_composite: OptimizeAttenuationComposite,
):
    set_mock_value(fake_composite.xspress3mini.channels[1].total_time, 100)
    set_mock_value(fake_composite.xspress3mini.channels[1].reset_ticks, 101)

    with patch(
        "mx_bluesky.hyperion.experiment_plans.optimise_attenuation_plan.is_deadtime_optimised",
        autospec=True,
    ) as mock_is_deadtime_optimised:
        RE(
            deadtime_optimisation(
                fake_composite,
                0.5,
                2,
                0.01,
                1,
                0.1,
                1e-6,
            )
        )
        mock_is_deadtime_optimised.assert_called_with(
            0.99, 0.01, 0.5, 0.1, Direction.POSITIVE
        )


@pytest.mark.parametrize(
    "target, upper_limit, lower_limit, default_high_roi, default_low_roi,initial_transmission,upper_transmission,lower_transmission",
    [
        (100, 90, 110, 1, 0, 0.5, 1, 0),
        (50, 100, 20, 10, 20, 0.5, 1, 0),
        (100, 100, 101, 10, 1, 0.5, 1, 0),
        (10, 100, 0, 2, 1, 0.5, 0, 1),
        (10, 100, 0, 2, 1, 0.5, 0.4, 0.1),
    ],
)
def test_check_parameters_fail_on_out_of_range_parameters(
    target: Literal[100] | Literal[50] | Literal[10],
    upper_limit: Literal[90] | Literal[100],
    lower_limit: Literal[110] | Literal[20] | Literal[101] | Literal[0],
    default_high_roi: Literal[1] | Literal[10] | Literal[2],
    default_low_roi: Literal[0] | Literal[20] | Literal[1],
    initial_transmission: float,
    upper_transmission: float | Literal[1] | Literal[0],
    lower_transmission: float | Literal[0] | Literal[1],
):
    with pytest.raises(ValueError):
        check_parameters(
            target,
            upper_limit,
            lower_limit,
            default_high_roi,
            default_low_roi,
            initial_transmission,
            upper_transmission,
            lower_transmission,
        )


def test_check_parameters_runs_on_correct_params():
    assert check_parameters(10, 100, 0, 2, 1, 0.5, 1, 0) is None


@pytest.mark.parametrize(
    "total_count, lower_limit, upper_limit",
    [(100, 99, 100), (100, 100, 100), (50, 25, 1000)],
)
def test_is_counts_within_target_is_true(
    total_count: Literal[100] | Literal[50],
    lower_limit: Literal[99] | Literal[100] | Literal[25],
    upper_limit: Literal[100] | Literal[1000],
):
    assert is_counts_within_target(total_count, lower_limit, upper_limit) is True


@pytest.mark.parametrize(
    "total_count, lower_limit, upper_limit",
    [(100, 101, 101), (0, 1, 2), (1000, 2000, 3000)],
)
def test_is_counts_within_target_is_false(
    total_count: Literal[100] | Literal[0] | Literal[1000],
    lower_limit: Literal[101] | Literal[1] | Literal[2000],
    upper_limit: Literal[101] | Literal[2] | Literal[3000],
):
    assert is_counts_within_target(total_count, lower_limit, upper_limit) is False


def test_total_count_exception_raised_after_max_cycles_reached(
    RE: RunEngine, fake_composite_mocked_sets: OptimizeAttenuationComposite
):
    optimise_attenuation_plan.is_counts_within_target = MagicMock(return_value=False)
    set_mock_value(
        fake_composite_mocked_sets.xspress3mini.dt_corrected_latest_mca[1],
        np.array([1, 1, 1, 1, 1, 1]),
    )
    with pytest.raises(AttenuationOptimisationFailedException):
        RE(
            total_counts_optimisation(
                fake_composite_mocked_sets, 1, 0, 10, 0, 5, 2, 1, 0, 0
            )
        )


@pytest.mark.parametrize(
    "direction, transmission, increment, upper_limit, lower_limit, new_transmission",
    [
        (Direction.POSITIVE, 0.5, 2, 0.9, 1e-6, 0.9),
        (Direction.POSITIVE, 0.1, 2, 0.9, 1e-6, 0.2),
        (Direction.NEGATIVE, 0.8, 2, 0.9, 1e-6, 0.4),
    ],
)
def test_deadtime_calc_new_transmission_gets_correct_value(
    direction: Direction | Direction,
    transmission: float,
    increment: Literal[2],
    upper_limit: float,
    lower_limit: float,
    new_transmission: float,
):
    assert (
        deadtime_calc_new_transmission(
            direction, transmission, increment, upper_limit, lower_limit
        )
        == new_transmission
    )


def test_deadtime_calc_new_transmission_raises_error_on_low_ransmission():
    with pytest.raises(AttenuationOptimisationFailedException):
        deadtime_calc_new_transmission(Direction.NEGATIVE, 1e-6, 2, 1, 1e-6)


def test_total_count_calc_new_transmission_raises_error_on_low_ransmission(
    RE: RunEngine, fake_composite_mocked_sets: OptimizeAttenuationComposite
):
    set_mock_value(
        fake_composite_mocked_sets.xspress3mini.dt_corrected_latest_mca[1],
        np.array([1, 1, 1, 1, 1, 1]),
    )
    with pytest.raises(AttenuationOptimisationFailedException):
        RE(
            total_counts_optimisation(
                fake_composite_mocked_sets,
                1e-6,
                0,
                1,
                10,
                20,
                1,
                1,
                0.5,
                0.1,
            )
        )


def test_total_counts_gets_within_target(
    RE: RunEngine,
    fake_composite_mocked_sets: OptimizeAttenuationComposite,
):
    # For simplicity we just increase the data array each iteration. In reality it's the transmission value that affects the array
    def update_data(value):
        nonlocal iteration
        iteration += 1
        set_mock_value(
            fake_composite_mocked_sets.xspress3mini.dt_corrected_latest_mca[1],
            np.array(([50, 50, 50, 50, 50]) * iteration),
        )
        return AsyncStatus(asyncio.sleep(0))

    fake_composite_mocked_sets.attenuator.set = update_data
    iteration = 0

    RE(
        total_counts_optimisation(
            fake_composite_mocked_sets,
            transmission=1,
            low_roi=0,
            high_roi=4,
            lower_count_limit=1000,
            upper_count_limit=2000,
            target_count=1500,
            max_cycles=10,
            upper_transmission_limit=1,
            lower_transmission_limit=0,
        )
    )


@pytest.mark.parametrize(
    "optimisation_type",
    [("total_counts"), ("deadtime")],
)
@patch(
    "mx_bluesky.hyperion.experiment_plans.optimise_attenuation_plan.total_counts_optimisation",
    autospec=True,
)
@patch(
    "mx_bluesky.hyperion.experiment_plans.optimise_attenuation_plan.deadtime_optimisation",
    autospec=True,
)
@patch(
    "mx_bluesky.hyperion.experiment_plans.optimise_attenuation_plan.check_parameters",
    autospec=True,
)
def test_optimisation_attenuation_plan_runs_correct_functions(
    mock_check_parameters,
    mock_deadtime_optimisation,
    mock_total_counts_optimisation,
    optimisation_type: Literal["total_counts"] | Literal["deadtime"],
    RE: RunEngine,
    fake_composite: OptimizeAttenuationComposite,
):
    fake_composite.attenuator.set = MagicMock(return_value=get_good_status())
    fake_composite.xspress3mini.acquire_time.set = MagicMock(
        return_value=get_good_status()
    )

    RE(
        optimise_attenuation_plan.optimise_attenuation_plan(
            fake_composite,
            optimisation_type=optimisation_type,
        )
    )

    if optimisation_type == "total_counts":
        mock_deadtime_optimisation.assert_not_called()
        mock_total_counts_optimisation.assert_called_once()
    else:
        mock_total_counts_optimisation.assert_not_called()
        mock_deadtime_optimisation.assert_called_once()
    fake_composite.attenuator.set.assert_called_once()
    mock_check_parameters.assert_called_once()
    fake_composite.xspress3mini.acquire_time.set.assert_called_once()



from unittest.mock import patch

import pytest
from dodal.devices.i24.aperture import Aperture
from dodal.devices.i24.beam_center import DetectorBeamCenter
from dodal.devices.i24.beamstop import Beamstop
from dodal.devices.i24.dual_backlight import DualBacklight
from dodal.devices.i24.i24_detector_motion import DetectorMotion

from mx_bluesky.beamlines.i24.serial.setup_beamline import setup_beamline

from ..conftest import TEST_LUT


@patch("mx_bluesky.beamlines.i24.serial.setup_beamline.setup_beamline.bps.sleep")
async def test_setup_beamline_for_collection_plan(
    _, aperture: Aperture, backlight: DualBacklight, beamstop: Beamstop, RE
):
    RE(setup_beamline.setup_beamline_for_collection_plan(aperture, backlight, beamstop))

    assert await aperture.position.get_value() == "In"
    assert await beamstop.pos_select.get_value() == "Data Collection"
    assert await beamstop.y_rotation.user_setpoint.get_value() == 0

    assert await backlight.backlight_position.pos_level.get_value() == "Out"


async def test_move_detector_stage_to_position_plan(detector_stage: DetectorMotion, RE):
    det_dist = 100
    RE(setup_beamline.move_detector_stage_to_position_plan(detector_stage, det_dist))

    assert await detector_stage.z.user_setpoint.get_value() == det_dist


def test_compute_beam_center_position_from_lut(dummy_params_ex):
    lut_path = TEST_LUT[dummy_params_ex.detector_name]

    expected_beam_x = 1597.06
    expected_beam_y = 1693.33

    beam_center_pos = setup_beamline.compute_beam_center_position_from_lut(
        lut_path,
        dummy_params_ex.detector_distance_mm,
        dummy_params_ex.detector_size_constants,
    )
    assert beam_center_pos[0] == pytest.approx(expected_beam_x, 1e-2)
    assert beam_center_pos[1] == pytest.approx(expected_beam_y, 1e-2)


async def test_set_detector_beam_center_plan(
    eiger_beam_center: DetectorBeamCenter, dummy_params_ex, RE
):
    beam_center_pos = setup_beamline.compute_beam_center_position_from_lut(
        TEST_LUT[dummy_params_ex.detector_name],
        dummy_params_ex.detector_distance_mm,  # 100
        dummy_params_ex.detector_size_constants,
    )
    # test_detector_distance = 100
    # test_detector_params = dummy_params_ex.detector_params
    RE(
        setup_beamline.set_detector_beam_center_plan(
            eiger_beam_center,
            beam_center_pos,  # test_detector_params, test_detector_distance
        )
    )

    assert await eiger_beam_center.beam_x.get_value() == pytest.approx(1597.06, 1e-2)
    assert await eiger_beam_center.beam_y.get_value() == pytest.approx(1693.33, 1e-2)


@patch("mx_bluesky.beamlines.i24.serial.setup_beamline.setup_beamline.caput")
@patch("mx_bluesky.beamlines.i24.serial.setup_beamline.setup_beamline.caget")
def test_pilatus_raises_error_if_fastchip_and_no_args_list(fake_caget, fake_caput, RE):
    with pytest.raises(TypeError):
        RE(setup_beamline.pilatus("fastchip", None))


@patch("mx_bluesky.beamlines.i24.serial.setup_beamline.setup_beamline.caput")
@patch("mx_bluesky.beamlines.i24.serial.setup_beamline.setup_beamline.caget")
@patch("mx_bluesky.beamlines.i24.serial.setup_beamline.setup_beamline.bps.sleep")
def test_pilatus_quickshot(_, fake_caget, fake_caput, RE):
    RE(setup_beamline.pilatus("quickshot", ["", "", 1, 0.1]))
    assert fake_caput.call_count == 10
    assert fake_caget.call_count == 2


@patch("mx_bluesky.beamlines.i24.serial.setup_beamline.setup_beamline.caput")
@patch("mx_bluesky.beamlines.i24.serial.setup_beamline.setup_beamline.caget")
@patch("mx_bluesky.beamlines.i24.serial.setup_beamline.setup_beamline.bps.sleep")
def test_pilatus_fastchip(_, fake_caget, fake_caput, RE):
    RE(setup_beamline.pilatus("fastchip", ["", "", 10, 0.1]))
    assert fake_caput.call_count == 13
    assert fake_caget.call_count == 2


@patch("mx_bluesky.beamlines.i24.serial.setup_beamline.setup_beamline.caput")
@patch("mx_bluesky.beamlines.i24.serial.setup_beamline.setup_beamline.caget")
def test_eiger_raises_error_if_quickshot_and_no_args_list(fake_caget, fake_caput, RE):
    with pytest.raises(TypeError):
        RE(setup_beamline.eiger("quickshot", None))


@patch("mx_bluesky.beamlines.i24.serial.setup_beamline.setup_beamline.caput")
@patch("mx_bluesky.beamlines.i24.serial.setup_beamline.setup_beamline.caget")
@patch("mx_bluesky.beamlines.i24.serial.setup_beamline.setup_beamline.bps.sleep")
def test_eiger_quickshot(_, fake_caget, fake_caput, RE):
    RE(setup_beamline.eiger("quickshot", ["", "", "1", "0.1"]))
    assert fake_caput.call_count == 30


@patch("mx_bluesky.beamlines.i24.serial.setup_beamline.setup_beamline.caput")
@patch("mx_bluesky.beamlines.i24.serial.setup_beamline.setup_beamline.caget")
@patch("mx_bluesky.beamlines.i24.serial.setup_beamline.setup_beamline.bps.sleep")
def test_eiger_triggered(_, fake_caget, fake_caput, RE):
    RE(setup_beamline.eiger("triggered", ["", "", "10", "0.1"]))
    assert fake_caget.call_count == 4
    assert fake_caput.call_count == 30


@pytest.mark.parametrize(
    "action, expected_caputs, expected_sleeps",
    [
        ("Pin_hand_mount", 11, 0),
        ("Pin_rt_hand_mount", 11, 0),
        ("Pin_data_collection", 12, 2),
        ("Pin_rt_data_collection", 13, 2),
    ],
)
@patch("mx_bluesky.beamlines.i24.serial.setup_beamline.setup_beamline.caput")
@patch("mx_bluesky.beamlines.i24.serial.setup_beamline.setup_beamline.bps.sleep")
def test_mode_change(
    fake_sleep, fake_caput, action, expected_caputs, expected_sleeps, RE
):
    RE(setup_beamline.modechange(action))
    assert fake_caput.call_count == expected_caputs
    assert fake_sleep.call_count == expected_sleeps



"""
Deploy latest release of mx-bluesky either on a beamline or in dev mode.
"""

import argparse
import os
import re
import shutil
import subprocess
import sys
from typing import NamedTuple
from uuid import uuid1

from create_venv import run_process_and_print_output, setup_venv
from git import Repo
from packaging.version import VERSION_PATTERN, Version

usage = "python %(prog)s beamline [options]"

recognised_beamlines = ["i03", "i04", "i24", "i02-1"]

VERSION_PATTERN_COMPILED = re.compile(
    f"^{VERSION_PATTERN}$", re.VERBOSE | re.IGNORECASE
)

DEV_DEPLOY_LOCATION = "/scratch/30day_tmp/mx-bluesky_release_test/bluesky"

MAX_DEPLOYMENTS = 4

help_message = f"""
To deploy mx_bluesky on a specific beamline, using the control machine to create the \
environment and without kubernetes, only the beamline argument needs to be passed.
This will put the latest release in /dls_sw/ixx/software/bluesky/mx_bluesky_v#.#.# and \
set the permissions accordingly. \n
To run in dev mode, pass also the --dev option. This will place a test release in \
{DEV_DEPLOY_LOCATION}. \n
"""


class Options(NamedTuple):
    release_dir: str
    kubernetes: bool = False
    print_release_dir: bool = False
    quiet: bool = False
    dev_mode: bool = False
    prune_deployments: bool = True

    # NOTE For i24 for now will need to set it to false from the command line
    use_control_machine: bool = True


class Deployment:
    # Set name, setup remote origin, get the latest version"""
    def __init__(self, name: str, repo_args, options: Options):
        self.name = name
        self.repo = Repo(repo_args)

        self.origin = self.repo.remotes.origin
        self.origin.fetch()
        self.origin.fetch("refs/tags/*:refs/tags/*")

        self.versions = [
            t.name for t in self.repo.tags if VERSION_PATTERN_COMPILED.match(t.name)
        ]
        self.versions.sort(key=Version, reverse=True)

        self.options = options
        if not self.options.quiet:
            print(f"Found {self.name}_versions:\n{os.linesep.join(self.versions)}")

        self.latest_version_str = self.versions[0]

    def deploy(self, beamline: str):
        print(f"Cloning latest version {self.name} into {self.deploy_location}")

        deploy_repo = Repo.init(self.deploy_location)
        deploy_origin = deploy_repo.create_remote("origin", self.origin.url)

        deploy_origin.fetch()
        deploy_origin.fetch("refs/tags/*:refs/tags/*")
        deploy_repo.git.checkout(self.latest_version_str)

        print("Setting permissions")
        groups_to_give_permission = _get_permission_groups(
            beamline, self.options.dev_mode
        )
        setfacl_params = ",".join(
            [f"g:{group}:rwx" for group in groups_to_give_permission]
        )

        # Set permissions and defaults
        os.system(f"setfacl -R -m {setfacl_params} {self.deploy_location}")
        os.system(f"setfacl -dR -m {setfacl_params} {self.deploy_location}")

    # Deploy location depends on the latest hyperion version (...software/bluesky/hyperion_V...)
    def set_deploy_location(self, release_area):
        self.deploy_location = os.path.join(release_area, self.name)
        if os.path.isdir(self.deploy_location):
            raise Exception(
                f"{self.deploy_location} already exists, stopping deployment for {self.name}"
            )


# Get permission groups depending on beamline/dev install
def _get_permission_groups(beamline: str, dev_mode: bool = False) -> list:
    beamline_groups = ["gda2", "dls_dasc"]
    if not dev_mode:
        beamline_groups.append(f"{beamline}_staff")
    return beamline_groups


def _create_environment_from_control_machine(
    mx_repo: Deployment,
    path_to_create_venv: str,
    path_to_dls_dev_env: str,
    beamline: str = "i03",
):
    try:
        user = os.environ["USER"]
    except KeyError:
        user = input(
            """Couldn't find username from the environment. Enter FedID in order \
                to SSH to control machine:"""
        )
    cmd = f"ssh {user}@{beamline}-control python3 {path_to_create_venv} {path_to_dls_dev_env} {mx_repo.deploy_location}"

    process = None
    try:
        # Call python script on i03-control to create the environment
        process = subprocess.Popen(cmd, shell=True, text=True, bufsize=1)
        process.communicate()
        if process.returncode != 0:
            print("Error occurred, exiting")
            sys.exit(1)
    except Exception as e:
        print(f"Exception while trying to install venv on i03-control: {e}")
    finally:
        if process:
            process.kill()


def _prune_old_deployments(release_area: str):
    def get_creation_time(deployment):
        # Warning: getctime gives time since last metadata change, not the creation time.
        return os.path.getctime(os.path.join(release_area, deployment))

    deployments = os.listdir(release_area)
    set_of_deployments = set(deployments)

    # Seperates symlinks and deployments
    symlinks = set()
    for item in deployments:
        if os.path.islink(os.path.join(release_area, item)):
            symlinks.add(item)
            set_of_deployments.remove(item)

    sorted_deployments = sorted(set_of_deployments, key=get_creation_time, reverse=True)

    # Excludes most recent deployments
    if len(sorted_deployments) > MAX_DEPLOYMENTS:
        full_path_old_deployments = {
            os.path.join(release_area, deployment)
            for deployment in sorted_deployments[MAX_DEPLOYMENTS:]
        }

        # Excludes deployments that are symlinked
        for link in symlinks:
            link_path = os.path.dirname(os.readlink(os.path.join(release_area, link)))
            if link_path in full_path_old_deployments:
                full_path_old_deployments.remove(link_path)

        # Deletes old deployments
        for old_deployment in full_path_old_deployments:
            print(f"Deleting old deployment {os.path.basename(old_deployment)}")
            shutil.rmtree(old_deployment)


def main(beamline: str, options: Options):
    release_area = options.release_dir
    this_repo_top = os.path.abspath(os.path.join(os.path.dirname(__file__), "../.."))

    if not options.quiet:
        print(f"Repo top is {this_repo_top}")

    mx_repo = Deployment(
        name="mx-bluesky",
        repo_args=os.path.join(this_repo_top, ".git"),
        options=options,
    )

    if mx_repo.name != "mx-bluesky":
        raise ValueError("This function should only be used with the mx-bluesky repo.")

    release_area_version = os.path.join(
        release_area, f"mx-bluesky_{mx_repo.latest_version_str}"
    )

    if options.print_release_dir:
        print(release_area_version)
        return

    print(f"Putting releases into {release_area_version}")

    dodal_repo = Deployment(
        name="dodal",
        repo_args=os.path.join(this_repo_top, "../dodal/.git"),
        options=options,
    )

    dodal_repo.set_deploy_location(release_area_version)
    mx_repo.set_deploy_location(release_area_version)

    # Deploy mx_bluesky repo
    mx_repo.deploy(beamline)

    # Now deploy the correct version of dodal
    dodal_repo.deploy(beamline)

    if not options.kubernetes:
        if mx_repo.name == "mx-bluesky":
            path_to_dls_dev_env = os.path.join(
                mx_repo.deploy_location,
                "utility_scripts/dls_dev_env.sh",
            )
            path_to_create_venv = os.path.join(
                mx_repo.deploy_location,
                "utility_scripts/deploy/create_venv.py",
            )

            # SSH into control machine if not in dev mode
            if options.use_control_machine and release_area != DEV_DEPLOY_LOCATION:
                _create_environment_from_control_machine(
                    mx_repo, path_to_create_venv, path_to_dls_dev_env, beamline
                )
            else:
                setup_venv(path_to_dls_dev_env, mx_repo.deploy_location)

    # If on beamline I24 also deploy the screens to run ssx collections
    if beamline == "i24":
        print("Setting up edm screens for serial collections on I24.")
        run_process_and_print_output("./utility_scripts/deploy/deploy_edm_for_ssx.sh")

    def create_symlink_by_tmp_and_rename(dirname, target, linkname):
        tmp_name = str(uuid1())
        target_path = os.path.join(dirname, target)
        linkname_path = os.path.join(dirname, linkname)
        tmp_path = os.path.join(dirname, tmp_name)
        os.symlink(target_path, tmp_path)
        os.rename(tmp_path, linkname_path)

    move_symlink = input(
        """Move symlink (y/n)? WARNING: this will affect the running version! \
            Only do so if you have informed the beamline scientist and you're sure \
            mx-bluesky is not running."""
    )
    if move_symlink == "y":
        # NOTE at some point need a better distinction between beamlines that use
        # hyperions and the rest of them. For now it's just i24 asaik.
        if beamline == "i24":
            live_location = os.path.join(release_area, "mx-bluesky")
            create_symlink_by_tmp_and_rename(release_area, live_location, "mx-bluesky")
        else:
            old_live_location = os.path.relpath(
                os.path.realpath(os.path.join(release_area, "hyperion")), release_area
            )
            make_live_stable_symlink = input(
                f"""The last live deployment was {old_live_location}, \
                    do you want to set this as the stable version? (y/n)"""
            )
            if make_live_stable_symlink == "y":
                create_symlink_by_tmp_and_rename(
                    release_area, old_live_location, "hyperion_stable"
                )

            relative_deploy_loc = os.path.join(
                os.path.relpath(mx_repo.deploy_location, release_area)
            )
            create_symlink_by_tmp_and_rename(
                release_area, relative_deploy_loc, "hyperion_latest"
            )
            create_symlink_by_tmp_and_rename(
                release_area, "hyperion_latest", "hyperion"
            )
        print(f"New version moved to {mx_repo.deploy_location}")
        print(
            """If running hyperion, to start this version run hyperion_restart \
                from the beamline's GDA"""
        )
    else:
        print("Quitting without latest version being updated")

    if options.prune_deployments:
        _prune_old_deployments(release_area)


# Get the release directory based off the beamline and the latest mx-bluesky version
def _parse_options() -> tuple[str, Options]:
    parser = argparse.ArgumentParser(
        formatter_class=argparse.RawTextHelpFormatter,
        description=__doc__,
        usage=usage,
        epilog=help_message,
    )
    parser.add_argument(
        "beamline",
        type=str,
        choices=recognised_beamlines,
        help="The beamline to deploy mx_bluesky to.",
    )
    parser.add_argument(
        "--dev",
        action="store_true",
        help=f"Test deployment in dev mode, install to {DEV_DEPLOY_LOCATION}.",
    )
    parser.add_argument(
        "--kubernetes",
        action="store_true",
        help="Prepare git workspaces for deployment to kubernetes; do not install virtual environment.",
    )
    parser.add_argument(
        "--print-release-dir",
        action="store_true",
        help="Print the path to the release folder and then exit.",
    )
    parser.add_argument(
        "-nc",
        "--no-control",
        action="store_false",
        help="Do not create environment running from the control machine.",
    )
    parser.add_argument(
        "-pd",
        "--prune-deployments",
        required=False,
        default=True,
        help="Delete deployments which are older than the latest four if they aren't being used in any symlinks",
    )

    args = parser.parse_args()
    if args.dev:
        print("Running as dev")
        release_dir = DEV_DEPLOY_LOCATION
    else:
        release_dir = f"/dls_sw/{args.beamline}/software/bluesky"

    return args.beamline, Options(
        release_dir=release_dir,
        kubernetes=args.kubernetes,
        print_release_dir=args.print_release_dir,
        quiet=args.print_release_dir,
        dev_mode=args.dev,
        prune_deployments=args.prune_deployments,
        use_control_machine=args.no_control,
    )


if __name__ == "__main__":
    beamline, options = _parse_options()
    main(beamline, options)



"""Interface for ``python -m mx_bluesky``."""

from argparse import ArgumentParser
from collections.abc import Sequence

from . import __version__

__all__ = ["main"]


def main(args: Sequence[str] | None = None) -> None:
    """Argument parser for the CLI."""
    parser = ArgumentParser()
    parser.add_argument(
        "-v",
        "--version",
        action="version",
        version=__version__,
    )
    parser.parse_args(args)


if __name__ == "__main__":
    main()



from unittest.mock import MagicMock, patch

import bluesky.plan_stubs as bps
import bluesky.preprocessors as bpp
import pytest
from bluesky.run_engine import RunEngine
from dodal.devices.oav.oav_detector import OAV
from dodal.devices.robot import BartRobot
from dodal.devices.webcam import Webcam
from ophyd_async.testing import set_mock_value

from mx_bluesky.common.external_interaction.ispyb.exp_eye_store import BLSampleStatus
from mx_bluesky.hyperion.external_interaction.callbacks.robot_load.ispyb_callback import (
    RobotLoadISPyBCallback,
)
from mx_bluesky.hyperion.parameters.constants import CONST

VISIT = "cm31105-4"

SAMPLE_ID = 231412
SAMPLE_PUCK = 50
SAMPLE_PIN = 4
ACTION_ID = 1098

metadata = {
    "subplan_name": CONST.PLAN.ROBOT_LOAD,
    "metadata": {
        "visit": VISIT,
        "sample_id": SAMPLE_ID,
        "sample_puck": SAMPLE_PUCK,
        "sample_pin": SAMPLE_PIN,
    },
    "activate_callbacks": [
        "RobotLoadISPyBCallback",
    ],
}


@patch(
    "mx_bluesky.hyperion.external_interaction.callbacks.robot_load.ispyb_callback.ExpeyeInteraction",
    autospec=True,
)
def test_given_start_doc_with_expected_data_then_data_put_in_ispyb(
    expeye: MagicMock,
):
    RE = RunEngine()
    RE.subscribe(RobotLoadISPyBCallback())
    expeye.return_value.start_load.return_value = ACTION_ID

    @bpp.run_decorator(md=metadata)
    def my_plan():
        yield from bps.null()

    RE(my_plan())

    expeye.return_value.start_load.assert_called_once_with(
        "cm31105", 4, SAMPLE_ID, SAMPLE_PUCK, SAMPLE_PIN
    )
    expeye.return_value.end_load.assert_called_once_with(ACTION_ID, "success", "OK")


@patch(
    "mx_bluesky.hyperion.external_interaction.callbacks.robot_load.ispyb_callback.ExpeyeInteraction",
    autospec=True,
)
def test_given_failing_plan_then_exception_detail(
    expeye: MagicMock,
):
    RE = RunEngine()
    RE.subscribe(RobotLoadISPyBCallback())
    expeye.return_value.start_load.return_value = ACTION_ID

    class _Exception(Exception): ...

    @bpp.run_decorator(md=metadata)
    def my_plan():
        raise _Exception("BAD")
        yield from bps.null()

    with pytest.raises(_Exception):
        RE(my_plan())

    expeye.return_value.start_load.assert_called_once_with(
        "cm31105", 4, SAMPLE_ID, SAMPLE_PUCK, SAMPLE_PIN
    )
    expeye.return_value.end_load.assert_called_once_with(ACTION_ID, "fail", "BAD")


@patch(
    "mx_bluesky.hyperion.external_interaction.callbacks.robot_load.ispyb_callback.ExpeyeInteraction.end_load"
)
@patch(
    "mx_bluesky.hyperion.external_interaction.callbacks.robot_load.ispyb_callback.ExpeyeInteraction"
    ".update_sample_status",
    new=MagicMock(),
)
def test_given_end_called_but_no_start_then_exception_raised(end_load):
    callback = RobotLoadISPyBCallback()
    callback.active = True
    with pytest.raises(AssertionError):
        callback.activity_gated_stop({"run_uid": None})  # type: ignore
    end_load.assert_not_called()


@bpp.run_decorator(md=metadata)
def successful_robot_load_plan(robot: BartRobot, oav: OAV, webcam: Webcam):
    yield from bps.create(name=CONST.DESCRIPTORS.ROBOT_LOAD)
    yield from bps.read(robot.barcode)
    yield from bps.read(oav.snapshot)
    yield from bps.read(webcam)
    yield from bps.save()


@bpp.run_decorator(md=metadata)
def unsuccessful_robot_load_plan():
    yield from []
    raise AssertionError("Test failure")


@patch(
    "mx_bluesky.hyperion.external_interaction.callbacks.robot_load.ispyb_callback.ExpeyeInteraction",
    autospec=True,
)
def test_given_plan_reads_barcode_then_data_put_in_ispyb(
    expeye: MagicMock,
    robot: BartRobot,
    oav: OAV,
    webcam: Webcam,
):
    RE = RunEngine()
    RE.subscribe(RobotLoadISPyBCallback())
    expeye.return_value.start_load.return_value = ACTION_ID

    set_mock_value(oav.snapshot.last_saved_path, "test_oav_snapshot")
    set_mock_value(webcam.last_saved_path, "test_webcam_snapshot")

    RE(successful_robot_load_plan(robot, oav, webcam))

    expeye.return_value.start_load.assert_called_once_with(
        "cm31105", 4, SAMPLE_ID, SAMPLE_PUCK, SAMPLE_PIN
    )
    expeye.return_value.update_barcode_and_snapshots.assert_called_once_with(
        ACTION_ID, "BARCODE", "test_webcam_snapshot", "test_oav_snapshot"
    )
    expeye.return_value.end_load.assert_called_once_with(ACTION_ID, "success", "OK")


@patch(
    "mx_bluesky.hyperion.external_interaction.callbacks.robot_load.ispyb_callback.ExpeyeInteraction",
    autospec=True,
)
def test_robot_load_complete_triggers_bl_sample_status_loaded(
    mock_sample_handling,
    RE: RunEngine,
    robot: BartRobot,
    oav: OAV,
    webcam: Webcam,
):
    RE.subscribe(RobotLoadISPyBCallback())

    RE(successful_robot_load_plan(robot, oav, webcam))

    mock_sample_handling.return_value.update_sample_status.assert_called_with(
        SAMPLE_ID, BLSampleStatus.LOADED
    )


@patch(
    "mx_bluesky.hyperion.external_interaction.callbacks.robot_load.ispyb_callback.ExpeyeInteraction",
    autospec=True,
)
def test_robot_load_fails_triggers_bl_sample_status_error(
    mock_sample_handling,
    RE: RunEngine,
    robot: BartRobot,
    oav: OAV,
    webcam: Webcam,
):
    RE.subscribe(RobotLoadISPyBCallback())

    with pytest.raises(AssertionError, match="Test failure"):
        RE(unsuccessful_robot_load_plan())

    mock_sample_handling.return_value.update_sample_status.assert_called_with(
        SAMPLE_ID, BLSampleStatus.ERROR_BEAMLINE
    )



from __future__ import annotations

import json

import bluesky.preprocessors as bpp
from blueapi.core import BlueskyContext
from bluesky.utils import MsgGenerator
from dodal.devices.eiger import EigerDetector
from dodal.devices.oav.oav_parameters import OAVParameters

from mx_bluesky.common.external_interaction.callbacks.xray_centre.ispyb_callback import (
    ispyb_activation_wrapper,
)
from mx_bluesky.common.parameters.constants import OavConstants
from mx_bluesky.common.utils.context import device_composite_from_context
from mx_bluesky.common.utils.log import LOGGER
from mx_bluesky.common.xrc_result import XRayCentreEventHandler
from mx_bluesky.hyperion.device_setup_plans.manipulate_sample import move_phi_chi_omega
from mx_bluesky.hyperion.device_setup_plans.utils import (
    start_preparing_data_collection_then_do_plan,
)
from mx_bluesky.hyperion.experiment_plans.change_aperture_then_move_plan import (
    change_aperture_then_move_to_xtal,
)
from mx_bluesky.hyperion.experiment_plans.grid_detect_then_xray_centre_plan import (
    GridDetectThenXRayCentreComposite,
    detect_grid_and_do_gridscan,
)
from mx_bluesky.hyperion.experiment_plans.oav_snapshot_plan import (
    setup_beamline_for_OAV,
)
from mx_bluesky.hyperion.experiment_plans.pin_tip_centring_plan import (
    PinTipCentringComposite,
    pin_tip_centre_plan,
)
from mx_bluesky.hyperion.parameters.constants import CONST
from mx_bluesky.hyperion.parameters.gridscan import (
    GridScanWithEdgeDetect,
    PinTipCentreThenXrayCentre,
)


def create_devices(context: BlueskyContext) -> GridDetectThenXRayCentreComposite:
    """
    GridDetectThenXRayCentreComposite contains all the devices we need, reuse that.
    """
    return device_composite_from_context(context, GridDetectThenXRayCentreComposite)


def create_parameters_for_grid_detection(
    pin_centre_parameters: PinTipCentreThenXrayCentre,
) -> GridScanWithEdgeDetect:
    params_json = json.loads(pin_centre_parameters.model_dump_json())
    del params_json["tip_offset_um"]
    grid_detect_and_xray_centre = GridScanWithEdgeDetect(**params_json)
    LOGGER.info(
        f"Parameters for grid detect and xray centre: {grid_detect_and_xray_centre.model_dump_json(indent=2)}"
    )
    return grid_detect_and_xray_centre


def pin_centre_then_flyscan_plan(
    composite: GridDetectThenXRayCentreComposite,
    parameters: PinTipCentreThenXrayCentre,
    oav_config_file: str = OavConstants.OAV_CONFIG_JSON,
):
    """Plan that performs a pin tip centre followed by a flyscan to determine the centres of interest"""

    pin_tip_centring_composite = PinTipCentringComposite(
        oav=composite.oav,
        smargon=composite.smargon,
        backlight=composite.backlight,
        pin_tip_detection=composite.pin_tip_detection,
    )

    def _pin_centre_then_flyscan_plan():
        yield from setup_beamline_for_OAV(
            composite.smargon, composite.backlight, composite.aperture_scatterguard
        )

        yield from move_phi_chi_omega(
            composite.smargon,
            parameters.phi_start_deg,
            parameters.chi_start_deg,
            group=CONST.WAIT.READY_FOR_OAV,
        )

        yield from pin_tip_centre_plan(
            pin_tip_centring_composite,
            parameters.tip_offset_um,
            oav_config_file,
        )

        grid_detect_params = create_parameters_for_grid_detection(parameters)

        oav_params = OAVParameters("xrayCentring", oav_config_file)

        yield from detect_grid_and_do_gridscan(
            composite,
            grid_detect_params,
            oav_params,
        )

    yield from ispyb_activation_wrapper(_pin_centre_then_flyscan_plan(), parameters)


def pin_tip_centre_then_xray_centre(
    composite: GridDetectThenXRayCentreComposite,
    parameters: PinTipCentreThenXrayCentre,
    oav_config_file: str = OavConstants.OAV_CONFIG_JSON,
) -> MsgGenerator:
    """Starts preparing for collection then performs the pin tip centre and xray centre"""
    eiger: EigerDetector = composite.eiger

    eiger.set_detector_parameters(parameters.detector_params)

    flyscan_event_handler = XRayCentreEventHandler()

    @bpp.subs_decorator(flyscan_event_handler)
    def pin_centre_flyscan_then_fetch_results() -> MsgGenerator:
        yield from start_preparing_data_collection_then_do_plan(
            composite.beamstop,
            eiger,
            composite.detector_motion,
            parameters.detector_params.detector_distance,
            pin_centre_then_flyscan_plan(composite, parameters, oav_config_file),
            group=CONST.WAIT.GRID_READY_FOR_DC,
        )

    yield from pin_centre_flyscan_then_fetch_results()
    flyscan_results = flyscan_event_handler.xray_centre_results
    assert flyscan_results, (
        "Flyscan result event not received or no crystal found and exception not raised"
    )
    yield from change_aperture_then_move_to_xtal(
        flyscan_results[0], composite.smargon, composite.aperture_scatterguard
    )



from unittest.mock import MagicMock, patch

import pytest
from bluesky.run_engine import RunEngine
from bluesky.simulators import RunEngineSimulator, assert_message_and_return_remaining
from bluesky.utils import Msg
from dodal.devices.i03.beamstop import BeamstopPositions
from dodal.devices.robot import SampleLocation

from mx_bluesky.hyperion.device_setup_plans.check_beamstop import BeamstopException
from mx_bluesky.hyperion.experiment_plans.flyscan_xray_centre_plan import (
    _fire_xray_centre_result_event,
)
from mx_bluesky.hyperion.experiment_plans.grid_detect_then_xray_centre_plan import (
    GridDetectThenXRayCentreComposite,
)
from mx_bluesky.hyperion.experiment_plans.robot_load_then_centre_plan import (
    RobotLoadThenCentreComposite,
    robot_load_then_xray_centre,
)
from mx_bluesky.hyperion.parameters.constants import CONST
from mx_bluesky.hyperion.parameters.gridscan import (
    PinTipCentreThenXrayCentre,
)
from mx_bluesky.hyperion.parameters.robot_load import RobotLoadThenCentre

from ....conftest import assert_none_matching, raw_params_from_file
from .conftest import FLYSCAN_RESULT_LOW, FLYSCAN_RESULT_MED, sim_fire_event_on_open_run


@pytest.fixture
def robot_load_then_centre_params():
    params = raw_params_from_file(
        "tests/test_data/parameter_json_files/good_test_robot_load_and_centre_params.json"
    )
    return RobotLoadThenCentre(**params)


@pytest.fixture
def robot_load_then_centre_params_no_energy(robot_load_then_centre_params):
    robot_load_then_centre_params.demand_energy_ev = None
    return robot_load_then_centre_params


@pytest.fixture
def sample_is_loaded(sim_run_engine, robot_load_then_centre_params):
    sample_location = SampleLocation(2, 6)
    robot_load_then_centre_params.sample_puck = sample_location.puck
    robot_load_then_centre_params.sample_pin = sample_location.pin
    mock_current_sample(sim_run_engine, sample_location)


@pytest.fixture
def sample_is_not_loaded(sim_run_engine, sample_is_loaded):
    mock_current_sample(sim_run_engine, SampleLocation(1, 1))


def mock_pin_centre_then_flyscan_plan(_, __):
    yield from _fire_xray_centre_result_event([FLYSCAN_RESULT_MED, FLYSCAN_RESULT_LOW])


@patch(
    "mx_bluesky.hyperion.experiment_plans.robot_load_then_centre_plan.pin_centre_then_flyscan_plan",
    side_effect=mock_pin_centre_then_flyscan_plan,
)
@patch(
    "mx_bluesky.hyperion.experiment_plans.robot_load_then_centre_plan.robot_load_and_change_energy_plan",
    MagicMock(return_value=iter([])),
)
def test_when_plan_run_then_centring_plan_run_with_expected_parameters(
    mock_centring_plan: MagicMock,
    robot_load_composite: RobotLoadThenCentreComposite,
    robot_load_then_centre_params: RobotLoadThenCentre,
):
    RE = RunEngine()

    RE(robot_load_then_xray_centre(robot_load_composite, robot_load_then_centre_params))
    composite_passed = mock_centring_plan.call_args[0][0]
    params_passed: PinTipCentreThenXrayCentre = mock_centring_plan.call_args[0][1]

    for name, value in vars(composite_passed).items():
        assert value == getattr(robot_load_composite, name)

    for name in GridDetectThenXRayCentreComposite.__dataclass_fields__.keys():
        assert getattr(composite_passed, name), f"{name} not in composite"

    assert isinstance(params_passed, PinTipCentreThenXrayCentre)
    assert params_passed.detector_params.expected_energy_ev == 11100


@patch(
    "mx_bluesky.hyperion.experiment_plans.robot_load_then_centre_plan.pin_centre_then_flyscan_plan",
    side_effect=mock_pin_centre_then_flyscan_plan,
)
@patch(
    "mx_bluesky.hyperion.experiment_plans.robot_load_then_centre_plan.robot_load_and_change_energy_plan",
    MagicMock(return_value=iter([])),
)
def test_when_plan_run_with_requested_energy_specified_energy_set_on_eiger(
    mock_centring_plan: MagicMock,
    robot_load_composite: RobotLoadThenCentreComposite,
    robot_load_then_centre_params: RobotLoadThenCentre,
    sim_run_engine: RunEngineSimulator,
):
    robot_load_composite.eiger.set_detector_parameters = MagicMock()
    sim_run_engine.add_handler_for_callback_subscribes()
    sim_fire_event_on_open_run(sim_run_engine, CONST.PLAN.FLYSCAN_RESULTS)
    sim_run_engine.simulate_plan(
        robot_load_then_xray_centre(robot_load_composite, robot_load_then_centre_params)
    )
    det_params = robot_load_composite.eiger.set_detector_parameters.call_args[0][0]
    assert det_params.expected_energy_ev == 11100
    params_passed: PinTipCentreThenXrayCentre = mock_centring_plan.call_args[0][1]
    assert params_passed.detector_params.expected_energy_ev == 11100


@patch(
    "mx_bluesky.hyperion.experiment_plans.robot_load_then_centre_plan.pin_centre_then_flyscan_plan",
    MagicMock(side_effect=mock_pin_centre_then_flyscan_plan),
)
@patch(
    "mx_bluesky.hyperion.experiment_plans.robot_load_then_centre_plan.robot_load_and_change_energy_plan",
    MagicMock(return_value=iter([])),
)
def test_given_no_energy_supplied_when_robot_load_then_centre_current_energy_set_on_eiger(
    robot_load_composite: RobotLoadThenCentreComposite,
    robot_load_then_centre_params_no_energy: RobotLoadThenCentre,
    sim_run_engine: RunEngineSimulator,
):
    robot_load_composite.eiger.set_detector_parameters = MagicMock()
    sim_run_engine.add_handler_for_callback_subscribes()
    sim_fire_event_on_open_run(sim_run_engine, CONST.PLAN.FLYSCAN_RESULTS)
    sim_run_engine.add_handler(
        "locate",
        lambda msg: {"readback": 11.105},
        "dcm-energy_in_kev",
    )
    sim_run_engine.simulate_plan(
        robot_load_then_xray_centre(
            robot_load_composite,
            robot_load_then_centre_params_no_energy,
        )
    )
    det_params = robot_load_composite.eiger.set_detector_parameters.call_args[0][0]
    assert det_params.expected_energy_ev == 11105


def run_simulating_smargon_wait(
    robot_load_then_centre_params,
    robot_load_composite,
    total_disabled_reads,
    sim_run_engine: RunEngineSimulator,
):
    num_of_reads = 0

    def return_not_disabled_after_reads(_):
        nonlocal num_of_reads
        num_of_reads += 1
        return {"values": {"value": int(num_of_reads < total_disabled_reads)}}

    sim_run_engine.add_handler(
        "locate",
        lambda msg: {"readback": 11.105},
        "dcm-energy_in_kev",
    )
    sim_run_engine.add_handler(
        "read", return_not_disabled_after_reads, "smargon-disabled"
    )

    return sim_run_engine.simulate_plan(
        robot_load_then_xray_centre(robot_load_composite, robot_load_then_centre_params)
    )


def dummy_robot_load_plan(*args, **kwargs):
    return (yield Msg("robot_load"))


@patch(
    "mx_bluesky.hyperion.experiment_plans.robot_load_then_centre_plan.pin_centre_then_flyscan_plan",
    MagicMock(side_effect=mock_pin_centre_then_flyscan_plan),
)
@patch(
    "mx_bluesky.hyperion.experiment_plans.robot_load_then_centre_plan.robot_load_and_change_energy_plan",
    MagicMock(side_effect=dummy_robot_load_plan),
)
def test_when_plan_run_then_detector_arm_started_before_wait_on_robot_load(
    robot_load_composite: RobotLoadThenCentreComposite,
    robot_load_then_centre_params: RobotLoadThenCentre,
    sim_run_engine,
):
    sim_run_engine.add_handler_for_callback_subscribes()
    sim_fire_event_on_open_run(sim_run_engine, CONST.PLAN.FLYSCAN_RESULTS)
    messages = sim_run_engine.simulate_plan(
        robot_load_then_xray_centre(robot_load_composite, robot_load_then_centre_params)
    )
    messages = assert_message_and_return_remaining(
        messages, lambda msg: msg.command == "set" and msg.obj.name == "eiger_do_arm"
    )

    assert_message_and_return_remaining(
        messages,
        lambda msg: msg.command == "robot_load",
    )


def mock_current_sample(sim_run_engine: RunEngineSimulator, sample: SampleLocation):
    sim_run_engine.add_handler(
        "read",
        lambda msg: {"robot-current_puck": {"value": sample.puck}},
        "robot-current_puck",
    )
    sim_run_engine.add_handler(
        "read",
        lambda msg: {"robot-current_pin": {"value": sample.pin}},
        "robot-current_pin",
    )


@patch(
    "mx_bluesky.hyperion.experiment_plans.robot_load_then_centre_plan.pin_centre_then_flyscan_plan",
    MagicMock(return_value=iter([Msg("centre_plan")])),
)
def test_given_sample_already_loaded_and_chi_not_changed_when_robot_load_called_then_eiger_not_staged_and_centring_not_run(
    robot_load_composite: RobotLoadThenCentreComposite,
    robot_load_then_centre_params: RobotLoadThenCentre,
    sim_run_engine: RunEngineSimulator,
    sample_is_loaded,
):
    sim_run_engine.add_handler_for_callback_subscribes()
    sim_fire_event_on_open_run(sim_run_engine, CONST.PLAN.FLYSCAN_RESULTS)
    robot_load_then_centre_params.chi_start_deg = None

    messages = sim_run_engine.simulate_plan(
        robot_load_then_xray_centre(
            robot_load_composite,
            robot_load_then_centre_params,
        )
    )

    assert_none_matching(
        messages, lambda msg: msg.command == "set" and msg.obj.name == "eiger_do_arm"
    )

    assert_none_matching(
        messages, lambda msg: msg.command == "set" and msg.obj.name == "robot"
    )

    assert_none_matching(
        messages,
        lambda msg: msg.command == "centre_plan",
    )


@patch(
    "mx_bluesky.hyperion.experiment_plans.robot_load_then_centre_plan.pin_centre_then_flyscan_plan",
    MagicMock(return_value=iter([Msg("centre_plan")])),
)
def test_given_sample_already_loaded_and_chi_is_changed_when_robot_load_called_then_eiger_staged_and_centring_run(
    robot_load_composite: RobotLoadThenCentreComposite,
    robot_load_then_centre_params: RobotLoadThenCentre,
    sim_run_engine: RunEngineSimulator,
    sample_is_loaded,
):
    sim_run_engine.add_handler_for_callback_subscribes()
    sim_fire_event_on_open_run(sim_run_engine, CONST.PLAN.FLYSCAN_RESULTS)
    robot_load_then_centre_params.chi_start_deg = 30

    messages = sim_run_engine.simulate_plan(
        robot_load_then_xray_centre(
            robot_load_composite,
            robot_load_then_centre_params,
        )
    )

    messages = assert_message_and_return_remaining(
        messages,
        lambda msg: msg.command == "set" and msg.obj.name == "eiger_do_arm",
    )

    assert_none_matching(
        messages, lambda msg: msg.command == "set" and msg.obj.name == "robot"
    )

    messages = assert_message_and_return_remaining(
        messages,
        lambda msg: msg.command == "centre_plan",
    )


@patch(
    "mx_bluesky.hyperion.experiment_plans.robot_load_then_centre_plan.pin_centre_then_flyscan_plan",
    MagicMock(return_value=iter([Msg("centre_plan")])),
)
@patch(
    "mx_bluesky.hyperion.experiment_plans.robot_load_then_centre_plan.robot_load_and_change_energy_plan",
    MagicMock(return_value=iter([Msg("full_robot_load_plan")])),
)
def test_given_sample_not_loaded_and_chi_not_changed_when_robot_load_called_then_eiger_staged_before_robot_and_centring_run_after(
    robot_load_composite: RobotLoadThenCentreComposite,
    robot_load_then_centre_params: RobotLoadThenCentre,
    sim_run_engine: RunEngineSimulator,
    sample_is_not_loaded,
):
    sim_run_engine.add_handler_for_callback_subscribes()
    sim_fire_event_on_open_run(sim_run_engine, CONST.PLAN.FLYSCAN_RESULTS)
    robot_load_then_centre_params.chi_start_deg = None

    messages = sim_run_engine.simulate_plan(
        robot_load_then_xray_centre(
            robot_load_composite,
            robot_load_then_centre_params,
        )
    )

    messages = assert_message_and_return_remaining(
        messages,
        lambda msg: msg.command == "set" and msg.obj.name == "eiger_do_arm",
    )

    messages = assert_message_and_return_remaining(
        messages, lambda msg: msg.command == "full_robot_load_plan"
    )

    messages = assert_message_and_return_remaining(
        messages,
        lambda msg: msg.command == "centre_plan",
    )


@patch(
    "mx_bluesky.hyperion.experiment_plans.robot_load_then_centre_plan.pin_centre_then_flyscan_plan",
    MagicMock(return_value=iter([Msg("centre_plan")])),
)
@patch(
    "mx_bluesky.hyperion.experiment_plans.robot_load_then_centre_plan.robot_load_and_change_energy_plan",
    MagicMock(return_value=iter([Msg("full_robot_load_plan")])),
)
def test_given_sample_not_loaded_and_chi_changed_when_robot_load_called_then_eiger_staged_before_robot_and_centring_run(
    robot_load_composite: RobotLoadThenCentreComposite,
    robot_load_then_centre_params: RobotLoadThenCentre,
    sim_run_engine: RunEngineSimulator,
    sample_is_not_loaded,
):
    sim_run_engine.add_handler_for_callback_subscribes()
    sim_fire_event_on_open_run(sim_run_engine, CONST.PLAN.FLYSCAN_RESULTS)
    robot_load_then_centre_params.chi_start_deg = 30

    messages = sim_run_engine.simulate_plan(
        robot_load_then_xray_centre(
            robot_load_composite,
            robot_load_then_centre_params,
        )
    )

    messages = assert_message_and_return_remaining(
        messages,
        lambda msg: msg.command == "set" and msg.obj.name == "eiger_do_arm",
    )

    messages = assert_message_and_return_remaining(
        messages, lambda msg: msg.command == "full_robot_load_plan"
    )

    messages = assert_message_and_return_remaining(
        messages,
        lambda msg: msg.command == "centre_plan",
    )


@patch(
    "mx_bluesky.hyperion.experiment_plans.robot_load_then_centre_plan.pin_centre_then_flyscan_plan",
    MagicMock(return_value=iter([Msg("centre_plan")])),
)
@patch(
    "mx_bluesky.hyperion.experiment_plans.robot_load_then_centre_plan.set_energy_plan",
    MagicMock(
        side_effect=lambda energy, _: iter([Msg("set_energy_plan", None, energy)])
    ),
)
def test_robot_load_then_centre_sets_energy_when_chi_change_and_no_robot_load(
    sim_run_engine: RunEngineSimulator,
    robot_load_composite: RobotLoadThenCentreComposite,
    robot_load_then_centre_params: RobotLoadThenCentre,
    sample_is_loaded,
):
    robot_load_then_centre_params.chi_start_deg = 30

    messages = sim_run_engine.simulate_plan(
        robot_load_then_xray_centre(
            robot_load_composite,
            robot_load_then_centre_params,
        )
    )

    messages = assert_message_and_return_remaining(
        messages, lambda msg: msg.command == "set_energy_plan" and msg.args[0] == 11100
    )
    messages = assert_message_and_return_remaining(
        messages, lambda msg: msg.command == "centre_plan"
    )


@patch(
    "mx_bluesky.hyperion.experiment_plans.robot_load_then_centre_plan.set_energy_plan",
    MagicMock(
        side_effect=lambda energy, _: iter([Msg("set_energy_plan", None, energy)])
    ),
)
def test_robot_load_then_centre_sets_energy_when_no_robot_load_no_chi_change(
    sim_run_engine: RunEngineSimulator,
    robot_load_composite: RobotLoadThenCentreComposite,
    robot_load_then_centre_params: RobotLoadThenCentre,
    sample_is_loaded,
):
    robot_load_then_centre_params.chi_start_deg = None

    messages = sim_run_engine.simulate_plan(
        robot_load_then_xray_centre(
            robot_load_composite,
            robot_load_then_centre_params,
        )
    )
    messages = assert_message_and_return_remaining(
        messages, lambda msg: msg.command == "set_energy_plan" and msg.args[0] == 11100
    )


def test_tip_offset_um_passed_to_pin_tip_centre_plan(
    robot_load_then_centre_params: RobotLoadThenCentre,
):
    robot_load_then_centre_params.tip_offset_um = 100
    assert (
        robot_load_then_centre_params.pin_centre_then_xray_centre_params.tip_offset_um
        == 100
    )


def test_robot_load_then_centre_fails_with_exception_when_no_beamstop(
    sim_run_engine: RunEngineSimulator,
    robot_load_composite: RobotLoadThenCentreComposite,
    robot_load_then_centre_params: RobotLoadThenCentre,
):
    sim_run_engine.add_read_handler_for(
        robot_load_composite.beamstop.selected_pos, BeamstopPositions.UNKNOWN
    )
    with pytest.raises(BeamstopException):
        sim_run_engine.simulate_plan(
            robot_load_then_xray_centre(
                robot_load_composite, robot_load_then_centre_params
            )
        )


@patch(
    "mx_bluesky.hyperion.experiment_plans.pin_centre_then_xray_centre_plan.detect_grid_and_do_gridscan"
)
def test_box_size_passed_through_to_gridscan(
    mock_detect_grid: MagicMock,
    robot_load_composite: RobotLoadThenCentreComposite,
    robot_load_then_centre_params: RobotLoadThenCentre,
    grid_detection_callback_with_detected_grid: MagicMock,
    RE: RunEngine,
):
    robot_load_then_centre_params.box_size_um = 25
    RE(
        robot_load_then_xray_centre(
            robot_load_composite,
            robot_load_then_centre_params,
        )
    )
    detect_grid_call = mock_detect_grid.mock_calls[0]
    assert detect_grid_call.args[1].box_size_um == 25



from pydantic import Field

from mx_bluesky.common.parameters.components import WithPandaGridScan
from mx_bluesky.hyperion.external_interaction.config_server import HyperionFeatureFlags


class WithHyperionUDCFeatures(WithPandaGridScan):
    features: HyperionFeatureFlags = Field(default=HyperionFeatureFlags())



import io
import json
import pickle
from unittest.mock import MagicMock, call, patch

import numpy as np
import pytest
from PIL import Image

from mx_bluesky.beamlines.i04.redis_to_murko_forwarder import (
    MURKO_ADDRESS,
    BatchMurkoForwarder,
    MurkoRequest,
    RedisListener,
    get_image_size,
    send_to_murko_and_get_results,
)


@pytest.fixture
def batch_forwarder():
    return BatchMurkoForwarder(redis_client=MagicMock(), batch_size=3)


@pytest.fixture
@patch("mx_bluesky.beamlines.i04.redis_to_murko_forwarder.StrictRedis")
def redis_listener(mock_redis):
    return RedisListener()


def test_given_image_added_when_flush_called_then_murko_called(
    batch_forwarder: BatchMurkoForwarder,
):
    batch_forwarder._handle_batch_of_images = MagicMock()
    batch_forwarder.add("sample_1", "uuid_1", np.zeros((256, 320)))
    batch_forwarder.flush()
    batch_forwarder._handle_batch_of_images.assert_called_once()


def test_given_no_image_added_when_flush_called_then_murko_not_called(
    batch_forwarder: BatchMurkoForwarder,
):
    batch_forwarder._handle_batch_of_images = MagicMock()
    batch_forwarder.flush()
    batch_forwarder._handle_batch_of_images.assert_not_called()


def test_given_already_flushed_when_flush_called_again_then_murko_not_called_twice(
    batch_forwarder: BatchMurkoForwarder,
):
    batch_forwarder._handle_batch_of_images = MagicMock()
    batch_forwarder.add("sample_1", "uuid_1", np.zeros((256, 320)))
    batch_forwarder.flush()
    batch_forwarder._handle_batch_of_images.assert_called_once()
    batch_forwarder._handle_batch_of_images.reset_mock()
    batch_forwarder.flush()
    batch_forwarder._handle_batch_of_images.assert_not_called()


def test_when_image_with_new_size_added_then_murko_called(
    batch_forwarder: BatchMurkoForwarder,
):
    batch_forwarder._handle_batch_of_images = MagicMock()
    batch_forwarder.add("sample_1", "uuid_1", np.zeros((256, 320)))
    batch_forwarder._handle_batch_of_images.assert_not_called()
    batch_forwarder.add("sample_1", "uuid_2", np.zeros((256, 450)))
    batch_forwarder._handle_batch_of_images.assert_called_once()


def test_when_more_images_added_than_batch_size_then_murko_called(
    batch_forwarder: BatchMurkoForwarder,
):
    batch_forwarder._handle_batch_of_images = MagicMock()
    for i in range(3):
        batch_forwarder.add("sample_1", f"uuid_{i}", np.zeros((256, 320)))

    batch_forwarder._handle_batch_of_images.assert_called_once()


def test_when_results_sent_to_redis_then_set_on_multiple_keys_but_published_once(
    batch_forwarder: BatchMurkoForwarder,
):
    results = [("uuid_1", {"result": 1}), ("uuid_2", {"result": 2})]
    batch_forwarder._send_murko_results_to_redis("sample_id", results)

    assert batch_forwarder.redis_client.hset.call_args_list == [  # type:ignore
        call(
            "murko:sample_id:results",
            "uuid_1",
            str(pickle.dumps({"result": 1})),
        ),
        call(
            "murko:sample_id:results",
            "uuid_2",
            str(pickle.dumps({"result": 2})),
        ),
    ]
    batch_forwarder.redis_client.publish.assert_called_once_with(  # type:ignore
        "murko-results", pickle.dumps(results)
    )


@patch(
    "mx_bluesky.beamlines.i04.redis_to_murko_forwarder.send_to_murko_and_get_results"
)
def test_when_images_flushed_then_results_are_gathered_correlated_and_sent_to_redis(
    mock_get_results: MagicMock, batch_forwarder: BatchMurkoForwarder
):
    mock_get_results.return_value = {
        "descriptions": [
            {"most_likely_click": (0, 1)},
            {"most_likely_click": (0.5, 0.75)},
        ]
    }
    batch_forwarder.add("sample_1", "uuid_1", np.zeros((256, 320)))
    batch_forwarder.add("sample_1", "uuid_2", np.zeros((256, 320)))
    batch_forwarder.flush()

    assert batch_forwarder.redis_client.hset.call_args_list == [  # type:ignore
        call(
            "murko:sample_1:results",
            "uuid_1",
            str(pickle.dumps({"most_likely_click": (0, 1)})),
        ),
        call(
            "murko:sample_1:results",
            "uuid_2",
            str(pickle.dumps({"most_likely_click": (0.5, 0.75)})),
        ),
    ]


def test_get_image_size_gives_expected_size():
    image = Image.new("RGB", (256, 512), color="white")
    image = np.asarray(image)
    width, height = get_image_size(image)
    assert width == 256
    assert height == 512


def test_when_listen_for_data_called_on_redis_listener_then_subscribes_to_expected_channel(
    redis_listener: RedisListener,
):
    # Raise exception to exit the infinite loop
    redis_listener._get_and_handle_message = MagicMock(side_effect=TimeoutError())

    with pytest.raises(TimeoutError):
        redis_listener.listen_for_image_data_forever()

    redis_listener.redis_client.pubsub().subscribe.assert_called_once_with("murko")  # type:ignore


def test_given_no_message_received_then_forwarder_flushed(
    redis_listener: RedisListener,
):
    redis_listener.forwarder = MagicMock()
    redis_listener.pubsub.get_message.return_value = None  # type:ignore
    redis_listener._get_and_handle_message()
    redis_listener.forwarder.flush.assert_called_once()


def get_jpeg_image():
    img = Image.new("RGB", (1, 1), color="black")
    img_byte_arr = io.BytesIO()
    img.save(img_byte_arr, format="JPEG")
    return img_byte_arr.getvalue()


def test_given_jpeg_image_received_then_data_retrieved_from_correct_redis_key(
    redis_listener: RedisListener,
):
    redis_listener.forwarder = MagicMock()
    data = {"uuid": "uuid_1", "sample_id": "sample_id_1"}
    redis_listener.pubsub.get_message.return_value = {  # type:ignore
        "type": "message",
        "data": json.dumps(data),
    }
    redis_listener.redis_client.hget.return_value = get_jpeg_image()  # type:ignore
    redis_listener._get_and_handle_message()
    redis_listener.redis_client.hget.assert_called_once_with(  # type:ignore
        "murko:sample_id_1:raw", "uuid_1"
    )


def test_given_jpeg_image_received_then_converted_to_numpy_array_and_sent_to_forwarder(
    redis_listener: RedisListener,
):
    redis_listener.forwarder = MagicMock()
    data = {"uuid": "uuid_1", "sample_id": "sample_id_1"}
    redis_listener.pubsub.get_message.return_value = {  # type:ignore
        "type": "message",
        "data": json.dumps(data),
    }
    redis_listener.redis_client.hget.return_value = get_jpeg_image()  # type:ignore
    redis_listener._get_and_handle_message()

    add_call = redis_listener.forwarder.add.call_args_list[0][0]  # type:ignore

    assert add_call[0] == "sample_id_1"
    assert add_call[1] == "uuid_1"
    assert np.array_equal(add_call[2], np.array([[[0, 0, 0]]]))


@patch("mx_bluesky.beamlines.i04.redis_to_murko_forwarder.zmq")
def test_send_to_murko_and_get_results_calls_murko_as_expected(patch_zmq):
    mock_request: MurkoRequest = {"prefix": "test"}  # type: ignore

    mock_socket = patch_zmq.Context.return_value.socket.return_value
    expected_return_dict = {"descriptions": ["returned"]}
    mock_socket.recv.return_value = pickle.dumps(expected_return_dict)

    returned = send_to_murko_and_get_results(mock_request)

    mock_socket.connect.assert_called_once_with(MURKO_ADDRESS)
    mock_socket.send.assert_called_once_with(pickle.dumps(mock_request))
    assert returned == expected_return_dict


@patch("mx_bluesky.beamlines.i04.redis_to_murko_forwarder.LOGGER")
def test_given_no_bytes_received_then_warn_and_do_nothing(
    patch_logger: MagicMock,
    redis_listener: RedisListener,
):
    redis_listener.forwarder = MagicMock()
    data = {"uuid": "uuid_1", "sample_id": "sample_id_1"}
    redis_listener.pubsub.get_message.return_value = {  # type:ignore
        "type": "message",
        "data": json.dumps(data),
    }
    redis_listener.redis_client.hget.return_value = None  # type:ignore
    redis_listener._get_and_handle_message()

    patch_logger.warning.assert_called_once()
    redis_listener.forwarder.add.assert_not_called()  # type:ignore



import asyncio
import time
from collections.abc import Callable
from types import ModuleType
from unittest.mock import MagicMock

import pytest
from bluesky.run_engine import RunEngine
from dodal.common.beamlines import beamline_parameters
from dodal.devices.zocalo import ZocaloTrigger
from dodal.utils import AnyDeviceFactory, collect_factories


@pytest.fixture
async def RE():
    RE = RunEngine(call_returns_result=True)
    # make sure the event loop is thoroughly up and running before we try to create
    # any ophyd_async devices which might need it
    timeout = time.monotonic() + 1
    while not RE.loop.is_running():
        await asyncio.sleep(0)
        if time.monotonic() > timeout:
            raise TimeoutError("This really shouldn't happen but just in case...")
    yield RE


MOCK_DAQ_CONFIG_PATH = "tests/devices/unit_tests/test_daq_configuration"
mock_paths = [
    ("DAQ_CONFIGURATION_PATH", MOCK_DAQ_CONFIG_PATH),
    ("ZOOM_PARAMS_FILE", "tests/devices/unit_tests/test_jCameraManZoomLevels.xml"),
    ("DISPLAY_CONFIG", "tests/devices/unit_tests/test_display.configuration"),
    ("LOOK_UPTABLE_DIR", "tests/devices/i10/lookupTables/"),
]
mock_attributes_table = {
    "i03": mock_paths,
    "i10": mock_paths,
    "s03": mock_paths,
    "i04": mock_paths,
    "s04": mock_paths,
    "i24": mock_paths,
}


def assert_event(mock_call, expected):
    actual = mock_call.args[0]
    if "data" in actual:
        actual = actual["data"]
    for k, v in expected.items():
        assert actual[k] == v, f"Mismatch in key {k}, {actual} <=> {expected}"


def mock_beamline_module_filepaths(bl_name, bl_module):
    if mock_attributes := mock_attributes_table.get(bl_name):
        [bl_module.__setattr__(attr[0], attr[1]) for attr in mock_attributes]
        beamline_parameters.BEAMLINE_PARAMETER_PATHS[bl_name] = (
            "tests/test_data/i04_beamlineParameters"
        )


def device_factories_for_beamline(beamline_module: ModuleType) -> set[AnyDeviceFactory]:
    return {
        f
        for f in collect_factories(beamline_module, include_skipped=True).values()
        if hasattr(f, "cache_clear")
    }


def modified_interactor_mock(assign_run_end: Callable | None = None):
    mock = MagicMock(spec=ZocaloTrigger)
    if assign_run_end:
        mock.run_end = assign_run_end
    return mock



from unittest.mock import MagicMock, patch

import pytest

from mx_bluesky.common.external_interaction.callbacks.xray_centre.ispyb_mapping import (
    construct_comment_for_gridscan,
)
from mx_bluesky.common.external_interaction.ispyb.data_model import (
    DataCollectionGridInfo,
    Orientation,
)
from mx_bluesky.hyperion.parameters.gridscan import HyperionSpecifiedThreeDGridScan

from .....conftest import (
    TEST_SAMPLE_ID,
    default_raw_gridscan_params,
)


@pytest.fixture
def dummy_params():
    dummy_params = HyperionSpecifiedThreeDGridScan(**default_raw_gridscan_params())
    dummy_params.sample_id = TEST_SAMPLE_ID
    dummy_params.run_number = 0
    return dummy_params


@patch("ispyb.open", autospec=True)
def test_ispyb_deposition_rounds_position_to_int(
    mock_ispyb_conn: MagicMock,
    dummy_params,
):
    assert construct_comment_for_gridscan(
        DataCollectionGridInfo(
            0.1,
            0.1,
            40,
            20,
            1.25,
            1.25,
            0.01,  # type: ignore
            100,
            Orientation.HORIZONTAL,
            True,  # type: ignore
        ),
    ) == (
        "Diffraction grid scan of 40 by 20 images "
        "in 100.0 um by 100.0 um steps. Top left (px): [0,100], bottom right (px): [3200,1700]."
    )


@pytest.mark.parametrize(
    ["raw", "rounded"],
    [
        (0.0012345, "1.2"),
        (0.020000000, "20.0"),
        (0.01999999, "20.0"),
        (0.015257, "15.3"),
        (0.0001234, "0.1"),
        (0.0017345, "1.7"),
        (0.0019945, "2.0"),
    ],
)
@patch(
    "mx_bluesky.common.external_interaction.callbacks.xray_centre.ispyb_mapping.oav_utils.bottom_right_from_top_left",
    autospec=True,
)
def test_ispyb_deposition_rounds_box_size_int(
    bottom_right_from_top_left: MagicMock,
    dummy_params: HyperionSpecifiedThreeDGridScan,
    raw,
    rounded,
):
    data_collection_grid_info = DataCollectionGridInfo(
        raw, raw, 0, 0, 1.25, 1.25, 0, 0, Orientation.HORIZONTAL, True
    )
    bottom_right_from_top_left.return_value = [0, 0]

    assert construct_comment_for_gridscan(data_collection_grid_info) == (
        "Diffraction grid scan of 0 by 0 images in "
        f"{rounded} um by {rounded} um steps. Top left (px): [0,0], bottom right (px): [0,0]."
    )



"""
Define beamline parameters for I03, Eiger detector and give an example of writing a
gridscan.
"""

from __future__ import annotations

import math
from pathlib import Path

from dodal.utils import get_beamline_name
from nexgen.nxs_utils import Attenuator, Beam, Detector, Goniometer, Source
from nexgen.nxs_write.nxmx_writer import NXmxFileWriter
from numpy.typing import DTypeLike
from scanspec.core import AxesPoints

from mx_bluesky.common.external_interaction.nexus.nexus_utils import (
    AxisDirection,
    create_detector_parameters,
    create_goniometer_axes,
    get_start_and_predicted_end_time,
)
from mx_bluesky.common.parameters.components import DiffractionExperimentWithSample


class NexusWriter:
    def __init__(
        self,
        parameters: DiffractionExperimentWithSample,
        data_shape: tuple[int, int, int],
        scan_points: AxesPoints,
        *,
        run_number: int | None = None,
        omega_start_deg: float = 0,
        chi_start_deg: float = 0,
        phi_start_deg: float = 0,
        vds_start_index: int = 0,
        # override default values when there is more than one collection per
        # detector arming event:
        full_num_of_images: int | None = None,
        meta_data_run_number: int | None = None,
        axis_direction: AxisDirection = AxisDirection.NEGATIVE,
    ) -> None:
        self.beam: Beam | None = None
        self.attenuator: Attenuator | None = None
        self.scan_points: dict = scan_points
        self.data_shape: tuple[int, int, int] = data_shape
        self.run_number: int = (
            run_number if run_number else parameters.detector_params.run_number
        )
        self.detector: Detector = create_detector_parameters(parameters.detector_params)
        self.source: Source = Source(get_beamline_name("S03"))
        self.directory: Path = Path(parameters.storage_directory)
        self.start_index: int = vds_start_index
        self.full_num_of_images: int = full_num_of_images or parameters.num_images
        self.data_filename: str = (
            f"{parameters.file_name}_{meta_data_run_number}"
            if meta_data_run_number
            else parameters.detector_params.full_filename
        )
        self.nexus_file: Path = (
            self.directory / f"{parameters.file_name}_{self.run_number}.nxs"
        )
        self.master_file: Path = (
            self.directory / f"{parameters.file_name}_{self.run_number}_master.h5"
        )
        self.goniometer: Goniometer = create_goniometer_axes(
            omega_start_deg,
            self.scan_points,
            chi=chi_start_deg,
            phi=phi_start_deg,
            omega_axis_direction=axis_direction,
        )

    def create_nexus_file(self, bit_depth: DTypeLike):
        """
        Creates a nexus file based on the parameters supplied when this object was
        initialised.
        """
        start_time, est_end_time = get_start_and_predicted_end_time(
            self.detector.exp_time * self.full_num_of_images
        )

        assert self.beam is not None
        assert self.attenuator is not None

        vds_shape = self.data_shape

        for filename in [self.nexus_file, self.master_file]:
            NXmx_Writer = NXmxFileWriter(
                filename,
                self.goniometer,
                self.detector,
                self.source,
                self.beam,
                self.attenuator,
                self.full_num_of_images,
            )
            NXmx_Writer.write(
                image_filename=f"{self.data_filename}",
                start_time=start_time,
                est_end_time=est_end_time,
            )
            NXmx_Writer.write_vds(
                vds_offset=self.start_index, vds_shape=vds_shape, vds_dtype=bit_depth
            )

    def get_image_datafiles(self, max_images_per_file=1000):
        return [
            self.directory / f"{self.data_filename}_{h5_num + 1:06}.h5"
            for h5_num in range(
                math.ceil(self.full_num_of_images / max_images_per_file)
            )
        ]



from __future__ import annotations

from collections.abc import Callable
from logging import Logger
from typing import TYPE_CHECKING, Any

from bluesky.callbacks import CallbackBase

if TYPE_CHECKING:
    from event_model.documents import Event, EventDescriptor, RunStart, RunStop


class PlanReactiveCallback(CallbackBase):
    log: Logger  # type: ignore # this is initialised to None and not annotated in the superclass

    def __init__(
        self,
        log: Logger,
        *,
        emit: Callable[..., Any] | None = None,
    ) -> None:
        """A callback base class which can be left permanently subscribed to a plan, and
        will 'activate' and 'deactivate' at the start and end of a plan which provides
        metadata to trigger this.
        The run_decorator of the plan should include in its metadata dictionary the key
        'activate callbacks', with a list of strings of the callback class(es) to
        activate or deactivate. On a recieving a start doc which specifies this, this
        class will be activated, and on recieving the stop document for the
        corresponding uid it will deactivate. The ordinary 'start', 'descriptor',
        'event' and 'stop' methods will be triggered as normal, and will in turn trigger
        'activity_gated_' methods - to preserve this functionality, subclasses which
        override 'start' etc. should include a call to super().start(...) etc.
        The logic of how activation is triggered will change to a more readable, version
        in the future (https://github.com/DiamondLightSource/hyperion/issues/964)."""

        super().__init__(emit=emit)
        self.emit_cb = emit  # to avoid GC; base class only holds a WeakRef
        self.active = False
        self.activity_uid = ""
        self.log = log

    def _run_activity_gated(self, name: str, func, doc, override=False):
        # Runs `func` if self.active is True or overide is true. Override can be used
        # to run the function even after setting self.active to False, i.e. in the last
        # handler of a run.

        running_gated_function = override or self.active
        if not running_gated_function:
            return doc
        try:
            return self.emit(name, func(doc))
        except Exception as e:
            self.log.exception(e)
            raise

    def start(self, doc: RunStart) -> RunStart | None:
        callbacks_to_activate = doc.get("activate_callbacks")
        if callbacks_to_activate and not self.active:
            activate = type(self).__name__ in callbacks_to_activate
            self.active = activate
            self.log.info(
                f"{'' if activate else 'not'} activating {type(self).__name__}"
            )
            self.activity_uid = doc.get("uid")
        return self._run_activity_gated("start", self.activity_gated_start, doc)

    def descriptor(self, doc: EventDescriptor) -> EventDescriptor | None:
        return self._run_activity_gated(
            "descriptor", self.activity_gated_descriptor, doc
        )

    def event(self, doc: Event) -> Event | None:  # type: ignore
        return self._run_activity_gated("event", self.activity_gated_event, doc)

    def stop(self, doc: RunStop) -> RunStop | None:
        do_stop = self.active
        if doc.get("run_start") == self.activity_uid:
            self.active = False
            self.activity_uid = ""
        return (
            self._run_activity_gated(
                "stop", self.activity_gated_stop, doc, override=True
            )
            if do_stop
            else doc
        )

    def activity_gated_start(self, doc: RunStart) -> RunStart | None:
        return doc

    def activity_gated_descriptor(self, doc: EventDescriptor) -> EventDescriptor | None:
        return doc

    def activity_gated_event(self, doc: Event) -> Event | None:
        return doc

    def activity_gated_stop(self, doc: RunStop) -> RunStop | None:
        return doc

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} with id: {hex(id(self))} - active: {self.active}>"



from unittest.mock import ANY, patch

import bluesky.plan_stubs as bps

from mx_bluesky.beamlines.i24.serial.web_gui_plans.general_plans import (
    gui_gonio_move_on_click,
    gui_move_detector,
    gui_sleep,
    gui_stage_move_on_click,
)


@patch("mx_bluesky.beamlines.i24.serial.web_gui_plans.general_plans.bps.sleep")
def test_gui_sleep(fake_sleep, RE):
    RE(gui_sleep(3))

    assert fake_sleep.call_count == 3


@patch("mx_bluesky.beamlines.i24.serial.web_gui_plans.general_plans.caput")
@patch(
    "mx_bluesky.beamlines.i24.serial.web_gui_plans.general_plans._move_detector_stage"
)
def test_gui_move_detector(fake_move_plan, fake_caput, RE):
    with patch(
        "mx_bluesky.beamlines.i24.serial.web_gui_plans.general_plans.i24.detector_motion"
    ):
        RE(gui_move_detector("eiger"))
        fake_move_plan.assert_called_once_with(ANY, -22.0)
    fake_caput.assert_called_once_with("ME14E-MO-IOC-01:GP101", "eiger")


@patch("mx_bluesky.beamlines.i24.serial.web_gui_plans.general_plans.bps.rd")
@patch("mx_bluesky.beamlines.i24.serial.web_gui_plans.general_plans.bps.mv")
def test_gui_gonio_move_on_click(fake_mv, fake_rd, RE):
    def fake_generator(value):
        yield from bps.null()
        return value

    fake_rd.side_effect = [fake_generator(1.25), fake_generator(1.25)]

    with (
        patch("mx_bluesky.beamlines.i24.serial.web_gui_plans.general_plans.i24.oav"),
        patch("mx_bluesky.beamlines.i24.serial.web_gui_plans.general_plans.i24.vgonio"),
    ):
        RE(gui_gonio_move_on_click((10, 20)))

    fake_mv.assert_called_with(ANY, 0.0125, ANY, 0.025)


@patch(
    "mx_bluesky.beamlines.i24.serial.web_gui_plans.general_plans._move_on_mouse_click_plan"
)
def test_gui_stage_move_on_click(fake_move_plan, RE):
    with (
        patch("mx_bluesky.beamlines.i24.serial.web_gui_plans.general_plans.i24.oav"),
        patch("mx_bluesky.beamlines.i24.serial.web_gui_plans.general_plans.i24.pmac"),
    ):
        RE(gui_stage_move_on_click((200, 200)))
        fake_move_plan.assert_called_once()



from unittest.mock import MagicMock, patch

import pytest

from mx_bluesky.common.external_interaction.ispyb.data_model import (
    DataCollectionInfo,
    DataCollectionPositionInfo,
    ScanDataInfo,
)
from mx_bluesky.common.external_interaction.ispyb.ispyb_store import (
    IspybIds,
    StoreInIspyb,
)
from mx_bluesky.hyperion.parameters.constants import CONST

from ......conftest import (
    EXPECTED_END_TIME,
    EXPECTED_START_TIME,
    TEST_BARCODE,
    TEST_DATA_COLLECTION_GROUP_ID,
    TEST_DATA_COLLECTION_IDS,
    TEST_SAMPLE_ID,
    TEST_SESSION_ID,
    assert_upsert_call_with,
    mx_acquisition_from_conn,
)

EXPECTED_UPDATE_DATA_COLLECTION = {
    "visitid": TEST_SESSION_ID,
    "parentid": TEST_DATA_COLLECTION_GROUP_ID,
    "sampleid": None,
    "detectorid": 78,
    "axisstart": 0.0,
    "axisrange": 0.1,
    "axisend": 180,
    "focal_spot_size_at_samplex": 1.0,
    "focal_spot_size_at_sampley": 1.0,
    "beamsize_at_samplex": 1,
    "beamsize_at_sampley": 1,
    "transmission": 100.0,
    "data_collection_number": 0,
    "detector_distance": 100.0,
    "exp_time": 0.1,
    "imgdir": "/tmp/",
    "imgprefix": "file_name",
    "imgsuffix": "h5",
    "n_passes": 1,
    "overlap": 0,
    "omegastart": 0,
    "start_image_number": 1,
    "wavelength": 123.98419840550369,
    "xbeam": 150.0,
    "ybeam": 160.0,
    "xtal_snapshot1": "test_1_y",
    "xtal_snapshot2": "test_2_y",
    "xtal_snapshot3": "test_3_y",
    "synchrotron_mode": None,
    "starttime": EXPECTED_START_TIME,
    "filetemplate": "file_name_1_master.h5",
    "nimages": 1800,
    "kappastart": 0,
}

EXPECTED_BEGIN_DATA_COLLECTION = EXPECTED_UPDATE_DATA_COLLECTION | {
    "comments": "Hyperion rotation scan",
}


@pytest.fixture
@patch(
    "mx_bluesky.common.external_interaction.callbacks.common.ispyb_mapping.get_current_time_string",
    new=MagicMock(return_value=EXPECTED_START_TIME),
)
def scan_data_info_for_begin():
    return ScanDataInfo(
        data_collection_info=DataCollectionInfo(
            omega_start=0.0,
            data_collection_number=0,
            xtal_snapshot1="test_1_y",
            xtal_snapshot2="test_2_y",
            xtal_snapshot3="test_3_y",
            n_images=1800,
            axis_range=0.1,
            axis_end=180.0,
            kappa_start=0.0,
            parent_id=None,
            visit_string="cm31105-4",
            sample_id=364758,
            detector_id=78,
            axis_start=0.0,
            focal_spot_size_at_samplex=1.0,
            focal_spot_size_at_sampley=1.0,
            beamsize_at_samplex=1.0,
            beamsize_at_sampley=1.0,
            transmission=100.0,
            comments="Hyperion rotation scan",
            detector_distance=100.0,
            exp_time=0.1,
            imgdir="/tmp/",
            file_template="file_name_1_master.h5",
            imgprefix="file_name",
            imgsuffix="h5",
            n_passes=1,
            overlap=0,
            start_image_number=1,
            wavelength=123.98419840550369,
            xbeam=150.0,
            ybeam=160.0,
            synchrotron_mode=None,
            undulator_gap1=None,
            start_time="2024-02-08 14:03:59",
        ),
        data_collection_id=None,
        data_collection_position_info=None,
        data_collection_grid_info=None,
    )


@pytest.fixture
def scan_data_info_for_update(scan_data_info_for_begin):
    return ScanDataInfo(
        data_collection_info=DataCollectionInfo(
            omega_start=0.0,
            data_collection_number=0,
            xtal_snapshot1="test_1_y",
            xtal_snapshot2="test_2_y",
            xtal_snapshot3="test_3_y",
            n_images=1800,
            axis_range=0.1,
            axis_end=180.0,
            kappa_start=0.0,
            parent_id=None,
            visit_string="cm31105-4",
            detector_id=78,
            axis_start=0.0,
            focal_spot_size_at_samplex=1.0,
            focal_spot_size_at_sampley=1.0,
            slitgap_vertical=1.0,
            slitgap_horizontal=1.0,
            beamsize_at_samplex=1.0,
            beamsize_at_sampley=1.0,
            transmission=100.0,
            detector_distance=100.0,
            exp_time=0.1,
            imgdir="/tmp/",
            file_template="file_name_1_master.h5",
            imgprefix="file_name",
            imgsuffix="h5",
            n_passes=1,
            overlap=0,
            flux=10.0,
            start_image_number=1,
            wavelength=123.98419840550369,
            xbeam=150.0,
            ybeam=160.0,
            synchrotron_mode="test",
            undulator_gap1=None,
            start_time="2024-02-08 14:03:59",
        ),
        data_collection_id=11,
        data_collection_position_info=DataCollectionPositionInfo(
            pos_x=10.0, pos_y=20.0, pos_z=30.0
        ),
        data_collection_grid_info=None,
    )


@pytest.fixture
def dummy_rotation_ispyb_with_experiment_type():
    store_in_ispyb = StoreInIspyb(CONST.SIM.ISPYB_CONFIG)
    return store_in_ispyb


@patch(
    "mx_bluesky.common.external_interaction.callbacks.common.ispyb_mapping.get_current_time_string",
    new=MagicMock(return_value=EXPECTED_START_TIME),
)
def test_begin_deposition(
    mock_ispyb_conn,
    dummy_rotation_ispyb,
    dummy_rotation_data_collection_group_info,
    scan_data_info_for_begin,
):
    assert scan_data_info_for_begin.data_collection_info.parent_id is None

    assert dummy_rotation_ispyb.begin_deposition(
        dummy_rotation_data_collection_group_info, [scan_data_info_for_begin]
    ) == IspybIds(
        data_collection_ids=(TEST_DATA_COLLECTION_IDS[0],),
        data_collection_group_id=TEST_DATA_COLLECTION_GROUP_ID,
    )

    assert (
        scan_data_info_for_begin.data_collection_info.parent_id
        == TEST_DATA_COLLECTION_GROUP_ID
    )

    mx_acq = mx_acquisition_from_conn(mock_ispyb_conn)
    assert_upsert_call_with(
        mx_acq.upsert_data_collection_group.mock_calls[0],
        mx_acq.get_data_collection_group_params(),
        {
            "parentid": TEST_SESSION_ID,
            "experimenttype": "SAD",
            "sampleid": TEST_SAMPLE_ID,
        },
    )
    assert_upsert_call_with(
        mx_acq.upsert_data_collection.mock_calls[0],
        mx_acq.get_data_collection_params(),
        EXPECTED_BEGIN_DATA_COLLECTION | {"sampleid": TEST_SAMPLE_ID},
    )


@patch(
    "mx_bluesky.common.external_interaction.callbacks.common.ispyb_mapping.get_current_time_string",
    new=MagicMock(return_value=EXPECTED_START_TIME),
)
def test_begin_deposition_with_group_id_updates_but_doesnt_insert(
    mock_ispyb_conn,
    dummy_rotation_data_collection_group_info,
    scan_data_info_for_begin,
):
    dummy_rotation_ispyb = StoreInIspyb(CONST.SIM.ISPYB_CONFIG)
    scan_data_info_for_begin.data_collection_info.parent_id = (
        TEST_DATA_COLLECTION_GROUP_ID
    )

    assert dummy_rotation_ispyb.begin_deposition(
        dummy_rotation_data_collection_group_info, [scan_data_info_for_begin]
    ) == IspybIds(
        data_collection_ids=(TEST_DATA_COLLECTION_IDS[0],),
        data_collection_group_id=TEST_DATA_COLLECTION_GROUP_ID,
    )
    mx_acq = mx_acquisition_from_conn(mock_ispyb_conn)
    assert_upsert_call_with(
        mx_acq.upsert_data_collection_group.mock_calls[0],
        mx_acq.get_data_collection_group_params(),
        {
            "id": TEST_DATA_COLLECTION_GROUP_ID,
            "parentid": TEST_SESSION_ID,
            "experimenttype": "SAD",
            "sampleid": TEST_SAMPLE_ID,
        },
    )
    assert (
        scan_data_info_for_begin.data_collection_info.parent_id
        == TEST_DATA_COLLECTION_GROUP_ID
    )

    assert_upsert_call_with(
        mx_acq.upsert_data_collection.mock_calls[0],
        mx_acq.get_data_collection_params(),
        EXPECTED_BEGIN_DATA_COLLECTION | {"sampleid": TEST_SAMPLE_ID},
    )


@patch(
    "mx_bluesky.common.external_interaction.callbacks.common.ispyb_mapping.get_current_time_string",
    new=MagicMock(return_value=EXPECTED_START_TIME),
)
def test_begin_deposition_with_alternate_experiment_type(
    mock_ispyb_conn,
    dummy_rotation_ispyb_with_experiment_type,
    dummy_rotation_data_collection_group_info,
    scan_data_info_for_begin,
):
    dummy_rotation_data_collection_group_info.experiment_type = "Characterization"
    assert dummy_rotation_ispyb_with_experiment_type.begin_deposition(
        dummy_rotation_data_collection_group_info,
        [scan_data_info_for_begin],
    ) == IspybIds(
        data_collection_ids=(TEST_DATA_COLLECTION_IDS[0],),
        data_collection_group_id=TEST_DATA_COLLECTION_GROUP_ID,
    )
    mx_acq = mx_acquisition_from_conn(mock_ispyb_conn)
    assert_upsert_call_with(
        mx_acq.upsert_data_collection_group.mock_calls[0],
        mx_acq.get_data_collection_group_params(),
        {
            "parentid": TEST_SESSION_ID,
            "experimenttype": "Characterization",
            "sampleid": TEST_SAMPLE_ID,
        },
    )


@patch(
    "mx_bluesky.common.external_interaction.ispyb.ispyb_store.get_current_time_string",
    new=MagicMock(return_value=EXPECTED_START_TIME),
)
def test_update_deposition(
    mock_ispyb_conn,
    dummy_rotation_ispyb,
    dummy_rotation_data_collection_group_info,
    scan_data_info_for_begin,
    scan_data_info_for_update,
):
    ispyb_ids = dummy_rotation_ispyb.begin_deposition(
        dummy_rotation_data_collection_group_info, [scan_data_info_for_begin]
    )
    mx_acq = mx_acquisition_from_conn(mock_ispyb_conn)
    mx_acq.upsert_data_collection_group.reset_mock()
    mx_acq.upsert_data_collection.reset_mock()

    scan_data_info_for_update.data_collection_info.parent_id = (
        ispyb_ids.data_collection_group_id
    )
    scan_data_info_for_update.data_collection_id = ispyb_ids.data_collection_ids[0]
    dummy_rotation_data_collection_group_info.sample_barcode = TEST_BARCODE

    assert dummy_rotation_ispyb.update_deposition(
        ispyb_ids,
        [scan_data_info_for_update],
    ) == IspybIds(
        data_collection_group_id=TEST_DATA_COLLECTION_GROUP_ID,
        data_collection_ids=(TEST_DATA_COLLECTION_IDS[0],),
    )
    mx_acq.upsert_data_collection_group.assert_not_called()
    assert_upsert_call_with(
        mx_acq.upsert_data_collection.mock_calls[0],
        mx_acq.get_data_collection_params(),
        EXPECTED_UPDATE_DATA_COLLECTION
        | {
            "id": TEST_DATA_COLLECTION_IDS[0],
            "synchrotron_mode": "test",
            "slitgap_vertical": 1,
            "slitgap_horizontal": 1,
            "flux": 10,
        },
    )

    assert_upsert_call_with(
        mx_acq.update_dc_position.mock_calls[0],
        mx_acq.get_dc_position_params(),
        {
            "id": TEST_DATA_COLLECTION_IDS[0],
            "pos_x": 10,
            "pos_y": 20,
            "pos_z": 30,
        },
    )


@patch(
    "mx_bluesky.common.external_interaction.callbacks.common.ispyb_mapping.get_current_time_string",
    new=MagicMock(return_value=EXPECTED_START_TIME),
)
def test_update_deposition_with_group_id_updates(
    mock_ispyb_conn,
    dummy_rotation_data_collection_group_info,
    scan_data_info_for_begin,
    scan_data_info_for_update,
):
    dummy_rotation_ispyb = StoreInIspyb(CONST.SIM.ISPYB_CONFIG)
    scan_data_info_for_begin.data_collection_info.parent_id = (
        TEST_DATA_COLLECTION_GROUP_ID
    )
    ispyb_ids = dummy_rotation_ispyb.begin_deposition(
        dummy_rotation_data_collection_group_info, [scan_data_info_for_begin]
    )
    mx_acq = mx_acquisition_from_conn(mock_ispyb_conn)
    mx_acq.upsert_data_collection_group.reset_mock()
    mx_acq.upsert_data_collection.reset_mock()

    scan_data_info_for_update.data_collection_info.parent_id = (
        ispyb_ids.data_collection_group_id
    )
    scan_data_info_for_update.data_collection_id = ispyb_ids.data_collection_ids[0]
    dummy_rotation_data_collection_group_info.sample_barcode = TEST_BARCODE
    assert dummy_rotation_ispyb.update_deposition(
        ispyb_ids,
        [scan_data_info_for_update],
    ) == IspybIds(
        data_collection_group_id=TEST_DATA_COLLECTION_GROUP_ID,
        data_collection_ids=(TEST_DATA_COLLECTION_IDS[0],),
    )
    mx_acq.upsert_data_collection_group.assert_not_called()
    assert_upsert_call_with(
        mx_acq.upsert_data_collection.mock_calls[0],
        mx_acq.get_data_collection_params(),
        EXPECTED_UPDATE_DATA_COLLECTION
        | {
            "id": TEST_DATA_COLLECTION_IDS[0],
            "synchrotron_mode": "test",
            "slitgap_vertical": 1,
            "slitgap_horizontal": 1,
            "flux": 10,
        },
    )

    assert_upsert_call_with(
        mx_acq.update_dc_position.mock_calls[0],
        mx_acq.get_dc_position_params(),
        {
            "id": TEST_DATA_COLLECTION_IDS[0],
            "pos_x": 10,
            "pos_y": 20,
            "pos_z": 30,
        },
    )


@patch(
    "mx_bluesky.common.external_interaction.callbacks.common.ispyb_mapping.get_current_time_string",
    new=MagicMock(return_value=EXPECTED_START_TIME),
)
@patch(
    "mx_bluesky.common.external_interaction.ispyb.ispyb_store.get_current_time_string",
)
def test_end_deposition_happy_path(
    get_current_time,
    mock_ispyb_conn,
    dummy_rotation_ispyb,
    dummy_rotation_data_collection_group_info,
    scan_data_info_for_begin,
    scan_data_info_for_update,
):
    ispyb_ids = dummy_rotation_ispyb.begin_deposition(
        dummy_rotation_data_collection_group_info, [scan_data_info_for_begin]
    )
    scan_data_info_for_update.data_collection_info.parent_id = (
        ispyb_ids.data_collection_group_id
    )
    scan_data_info_for_update.data_collection_id = ispyb_ids.data_collection_ids[0]
    ispyb_ids = dummy_rotation_ispyb.update_deposition(
        ispyb_ids,
        [scan_data_info_for_update],
    )
    mx_acq = mx_acquisition_from_conn(mock_ispyb_conn)
    mx_acq.upsert_data_collection_group.reset_mock()
    mx_acq.upsert_data_collection.reset_mock()
    mx_acq.upsert_dc_grid.reset_mock()

    get_current_time.return_value = EXPECTED_END_TIME
    dummy_rotation_ispyb.end_deposition(ispyb_ids, "success", "Test succeeded")
    assert mx_acq.update_data_collection_append_comments.call_args_list[0] == (
        (
            TEST_DATA_COLLECTION_IDS[0],
            "DataCollection Successful reason: Test succeeded",
            " ",
        ),
    )
    assert_upsert_call_with(
        mx_acq.upsert_data_collection.mock_calls[0],
        mx_acq.get_data_collection_params(),
        {
            "id": TEST_DATA_COLLECTION_IDS[0],
            "parentid": TEST_DATA_COLLECTION_GROUP_ID,
            "endtime": EXPECTED_END_TIME,
            "runstatus": "DataCollection Successful",
        },
    )
    assert len(mx_acq.upsert_data_collection.mock_calls) == 1


def test_store_rotation_scan_failures(
    mock_ispyb_conn, dummy_rotation_ispyb: StoreInIspyb
):
    ispyb_ids = IspybIds(
        data_collection_group_id=TEST_DATA_COLLECTION_GROUP_ID,
    )
    with pytest.raises(AssertionError):
        dummy_rotation_ispyb.end_deposition(ispyb_ids, "", "")


@pytest.mark.parametrize("dcgid", [2, 45, 61, 88, 13, 25])
def test_store_rotation_scan_uses_supplied_dcgid(
    mock_ispyb_conn,
    dcgid,
    dummy_rotation_data_collection_group_info,
    scan_data_info_for_begin,
    scan_data_info_for_update,
):
    mock_ispyb_conn.return_value.mx_acquisition.upsert_data_collection_group.return_value = dcgid
    store_in_ispyb = StoreInIspyb(CONST.SIM.ISPYB_CONFIG)
    scan_data_info_for_begin.data_collection_info.parent_id = dcgid
    ispyb_ids = store_in_ispyb.begin_deposition(
        dummy_rotation_data_collection_group_info, [scan_data_info_for_begin]
    )
    assert ispyb_ids.data_collection_group_id == dcgid
    mx = mx_acquisition_from_conn(mock_ispyb_conn)
    assert_upsert_call_with(
        mx.upsert_data_collection_group.mock_calls[0],
        mx.get_data_collection_group_params(),
        {
            "id": dcgid,
            "parentid": TEST_SESSION_ID,
            "experimenttype": "SAD",
            "sampleid": TEST_SAMPLE_ID,
        },
    )
    assert (
        store_in_ispyb.update_deposition(
            ispyb_ids,
            [scan_data_info_for_update],
        ).data_collection_group_id
        == dcgid
    )



"""
Define a mapping between the edm screens/IOC enum inputs for the general purpose PVs and
the map/chip/pump settings.
The enum values should not be changed unless they are also modified in the drop down
menu in the edm screen, as their order should always match.
New ones may be added if needed in the future.
"""

from enum import Enum, IntEnum


class MappingType(IntEnum):
    NoMap = 0
    Lite = 1

    def __str__(self) -> str:
        """Returns the mapping."""
        return self.name


# FIXME See https://github.com/DiamondLightSource/mx_bluesky/issues/77
class ChipType(IntEnum):
    Oxford = 0
    OxfordInner = 1
    Custom = 2
    Minichip = 3  # Mini oxford, 1 city block only
    MISP = 4  # New PSI polymer chip

    def __str__(self) -> str:
        """Returns the chip name."""
        return self.name


class PumpProbeSetting(IntEnum):
    NoPP = 0
    Short1 = 1
    Short2 = 2
    Repeat1 = 3
    Repeat2 = 4
    Repeat3 = 5
    Repeat5 = 6
    Repeat10 = 7
    Medium1 = 8

    def __str__(self) -> str:
        """Returns the pump-probe setting name."""
        return self.name


class Fiducials(str, Enum):
    origin = "origin"
    zero = "zero"
    fid1 = "f1"
    fid2 = "f2"



import json

from bluesky.callbacks import CallbackBase

from mx_bluesky.common.utils.log import LOGGER


class _BestEffortEncoder(json.JSONEncoder):
    def default(self, o):
        return repr(o)


def format_doc_for_log(doc):
    return json.dumps(doc, indent=2, cls=_BestEffortEncoder)


class VerbosePlanExecutionLoggingCallback(CallbackBase):
    def start(self, doc):
        LOGGER.info(f"START: {format_doc_for_log(doc)}")

    def descriptor(self, doc):
        LOGGER.info(f"DESCRIPTOR: {format_doc_for_log(doc)}")

    def event(self, doc):
        LOGGER.info(f"EVENT: {format_doc_for_log(doc)}")
        return doc

    def stop(self, doc):
        LOGGER.info(f"STOP: {format_doc_for_log(doc)}")



from unittest.mock import MagicMock, patch

from mx_bluesky.hyperion.external_interaction.callbacks.rotation.ispyb_callback import (
    RotationISPyBCallback,
)

from ......conftest import (
    EXPECTED_END_TIME,
    EXPECTED_START_TIME,
    TEST_DATA_COLLECTION_GROUP_ID,
    TEST_DATA_COLLECTION_IDS,
    TEST_SAMPLE_ID,
    TEST_SESSION_ID,
    TestData,
    assert_upsert_call_with,
    mx_acquisition_from_conn,
    remap_upsert_columns,
)

EXPECTED_DATA_COLLECTION = {
    "visitid": TEST_SESSION_ID,
    "parentid": TEST_DATA_COLLECTION_GROUP_ID,
    "sampleid": TEST_SAMPLE_ID,
    "detectorid": 78,
    "axisstart": 0.0,
    "axisrange": 0.1,
    "axisend": 180,
    "comments": "test",
    "data_collection_number": 1,
    "detector_distance": 100.0,
    "exp_time": 0.1,
    "imgdir": "/tmp/dls/i03/data/2024/cm31105-4/auto/123456/",
    "imgprefix": "file_name",
    "imgsuffix": "h5",
    "n_passes": 1,
    "overlap": 0,
    "omegastart": 0,
    "start_image_number": 1,
    "xbeam": 150.0,
    "ybeam": 160.0,
    "synchrotron_mode": None,
    "starttime": EXPECTED_START_TIME,
    "filetemplate": "file_name_1_master.h5",
    "nimages": 1800,
    "kappastart": None,
}


@patch(
    "mx_bluesky.common.external_interaction.callbacks.common.ispyb_mapping.get_current_time_string",
    new=MagicMock(return_value=EXPECTED_START_TIME),
)
def test_activity_gated_start(mock_ispyb_conn, test_rotation_start_outer_document):
    callback = RotationISPyBCallback()

    callback.activity_gated_start(test_rotation_start_outer_document)
    mx = mx_acquisition_from_conn(mock_ispyb_conn)
    assert_upsert_call_with(
        mx.upsert_data_collection_group.mock_calls[0],
        mx.get_data_collection_group_params(),
        {
            "parentid": TEST_SESSION_ID,
            "experimenttype": "SAD",
            "sampleid": TEST_SAMPLE_ID,
        },
    )
    assert_upsert_call_with(
        mx.upsert_data_collection.mock_calls[0],
        mx.get_data_collection_params(),
        EXPECTED_DATA_COLLECTION,
    )


@patch(
    "mx_bluesky.common.external_interaction.callbacks.common.ispyb_mapping.get_current_time_string",
    new=MagicMock(return_value=EXPECTED_START_TIME),
)
def test_hardware_read_events(mock_ispyb_conn, test_rotation_start_outer_document):
    callback = RotationISPyBCallback()
    callback.activity_gated_start(test_rotation_start_outer_document)  # pyright: ignore
    callback.activity_gated_start(
        TestData.test_rotation_start_main_document  # pyright: ignore
    )
    mx = mx_acquisition_from_conn(mock_ispyb_conn)

    mx.upsert_data_collection_group.reset_mock()
    mx.upsert_data_collection.reset_mock()

    callback.activity_gated_descriptor(
        TestData.test_descriptor_document_pre_data_collection
    )
    callback.activity_gated_event(TestData.test_event_document_pre_data_collection)
    mx.upsert_data_collection_group.assert_not_called()
    assert_upsert_call_with(
        mx.upsert_data_collection.mock_calls[0],
        mx.get_data_collection_params(),
        {
            "parentid": TEST_DATA_COLLECTION_GROUP_ID,
            "id": TEST_DATA_COLLECTION_IDS[0],
            "slitgaphorizontal": 0.1234,
            "slitgapvertical": 0.2345,
            "synchrotronmode": "User",
            "undulatorgap1": 1.234,
            "resolution": 1.1830593331191241,
            "wavelength": 1.11647184541378,
        },
    )
    mx.update_data_collection_append_comments.assert_called_with(
        TEST_DATA_COLLECTION_IDS[0], "Sample position (µm): (158, 24, 3)", " "
    )
    expected_data = TestData.test_event_document_pre_data_collection["data"]
    assert_upsert_call_with(
        mx.update_dc_position.mock_calls[0],
        mx.get_dc_position_params(),
        {
            "id": TEST_DATA_COLLECTION_IDS[0],
            "pos_x": expected_data["smargon-x"],
            "pos_y": expected_data["smargon-y"],
            "pos_z": expected_data["smargon-z"],
        },
    )


@patch(
    "mx_bluesky.common.external_interaction.callbacks.common.ispyb_mapping.get_current_time_string",
    new=MagicMock(return_value=EXPECTED_START_TIME),
)
def test_flux_read_events(mock_ispyb_conn, test_rotation_start_outer_document):
    callback = RotationISPyBCallback()
    callback.activity_gated_start(test_rotation_start_outer_document)  # pyright: ignore
    callback.activity_gated_start(
        TestData.test_rotation_start_main_document  # pyright: ignore
    )
    mx = mx_acquisition_from_conn(mock_ispyb_conn)
    callback.activity_gated_descriptor(
        TestData.test_descriptor_document_pre_data_collection
    )
    callback.activity_gated_event(TestData.test_event_document_pre_data_collection)
    mx.upsert_data_collection_group.reset_mock()
    mx.upsert_data_collection.reset_mock()
    callback.activity_gated_descriptor(
        TestData.test_descriptor_document_during_data_collection
    )
    callback.activity_gated_event(
        TestData.test_rotation_event_document_during_data_collection
    )

    mx.upsert_data_collection_group.assert_not_called()
    assert_upsert_call_with(
        mx.upsert_data_collection.mock_calls[0],
        mx.get_data_collection_params(),
        {
            "parentid": TEST_DATA_COLLECTION_GROUP_ID,
            "id": TEST_DATA_COLLECTION_IDS[0],
            "focal_spot_size_at_samplex": 0.05,
            "focal_spot_size_at_sampley": 0.02,
            "beamsize_at_samplex": 0.05,
            "beamsize_at_sampley": 0.02,
            "wavelength": 1.11647184541378,
            "transmission": 98,
            "flux": 9.81,
            "resolution": 1.1830593331191241,
        },
    )


@patch(
    "mx_bluesky.common.external_interaction.callbacks.common.ispyb_mapping.get_current_time_string",
    new=MagicMock(return_value=EXPECTED_START_TIME),
)
def test_oav_rotation_snapshot_triggered_event(
    mock_ispyb_conn, test_rotation_start_outer_document
):
    callback = RotationISPyBCallback()
    callback.activity_gated_start(test_rotation_start_outer_document)  # pyright: ignore
    callback.activity_gated_start(
        TestData.test_rotation_start_main_document  # pyright: ignore
    )
    mx = mx_acquisition_from_conn(mock_ispyb_conn)
    callback.activity_gated_descriptor(
        TestData.test_descriptor_document_oav_rotation_snapshot
    )

    for snapshot in [
        {"filename": "snapshot_0", "colname": "xtal_snapshot1"},
        {"filename": "snapshot_90", "colname": "xtal_snapshot2"},
        {"filename": "snapshot_180", "colname": "xtal_snapshot3"},
        {"filename": "snapshot_270", "colname": "xtal_snapshot4"},
    ]:
        mx.upsert_data_collection.reset_mock()
        event_doc = dict(TestData.test_event_document_oav_rotation_snapshot)
        event_doc["data"]["oav-snapshot-last_saved_path"] = snapshot["filename"]  # type: ignore
        callback.activity_gated_event(
            TestData.test_event_document_oav_rotation_snapshot
        )
        mx.upsert_data_collection_group.reset_mock()
        assert_upsert_call_with(
            mx.upsert_data_collection.mock_calls[0],
            mx.get_data_collection_params(),
            {
                "parentid": TEST_DATA_COLLECTION_GROUP_ID,
                "id": TEST_DATA_COLLECTION_IDS[0],
                snapshot["colname"]: snapshot["filename"],
            },
        )

    mx.upsert_data_collection_group.assert_not_called()


@patch(
    "mx_bluesky.common.external_interaction.callbacks.common.ispyb_mapping.get_current_time_string",
    new=MagicMock(return_value=EXPECTED_START_TIME),
)
def test_activity_gated_stop(mock_ispyb_conn, test_rotation_start_outer_document):
    callback = RotationISPyBCallback()
    callback.activity_gated_start(test_rotation_start_outer_document)  # pyright: ignore
    callback.activity_gated_start(
        TestData.test_rotation_start_main_document  # pyright: ignore
    )
    mx = mx_acquisition_from_conn(mock_ispyb_conn)

    mx.upsert_data_collection_group.reset_mock()
    mx.upsert_data_collection.reset_mock()

    with patch(
        "mx_bluesky.common.external_interaction.ispyb.ispyb_store.get_current_time_string",
        new=MagicMock(return_value=EXPECTED_END_TIME),
    ):
        callback.activity_gated_stop(TestData.test_rotation_stop_main_document)

    assert mx.update_data_collection_append_comments.call_args_list[0] == (
        (
            TEST_DATA_COLLECTION_IDS[0],
            "DataCollection Successful reason: Test succeeded",
            " ",
        ),
    )
    assert_upsert_call_with(
        mx.upsert_data_collection.mock_calls[0],
        mx.get_data_collection_params(),
        {
            "id": TEST_DATA_COLLECTION_IDS[0],
            "parentid": TEST_DATA_COLLECTION_GROUP_ID,
            "endtime": EXPECTED_END_TIME,
            "runstatus": "DataCollection Successful",
        },
    )
    assert len(mx.upsert_data_collection.mock_calls) == 1


def test_comment_correct_after_hardware_read(
    mock_ispyb_conn, test_rotation_start_outer_document
):
    callback = RotationISPyBCallback()
    test_rotation_start_outer_document["mx_bluesky_parameters"] = (
        test_rotation_start_outer_document["mx_bluesky_parameters"].replace(
            '"comment":"test"', '"comment":"a lovely unit test"'
        )
    )
    callback.activity_gated_start(test_rotation_start_outer_document)  # pyright: ignore
    callback.activity_gated_start(
        TestData.test_rotation_start_main_document  # pyright: ignore
    )
    mx = mx_acquisition_from_conn(mock_ispyb_conn)
    dc_upsert_dict = remap_upsert_columns(
        mx.get_data_collection_params(),
        mx.upsert_data_collection.mock_calls[0].args[0],
    )
    assert dc_upsert_dict["comments"] == "a lovely unit test"

    mx.upsert_data_collection_group.reset_mock()
    mx.upsert_data_collection.reset_mock()

    callback.activity_gated_descriptor(
        TestData.test_descriptor_document_pre_data_collection
    )
    callback.activity_gated_event(TestData.test_event_document_pre_data_collection)
    assert_upsert_call_with(
        mx.upsert_data_collection.mock_calls[0],
        mx.get_data_collection_params(),
        {
            "parentid": TEST_DATA_COLLECTION_GROUP_ID,
            "id": TEST_DATA_COLLECTION_IDS[0],
            "slitgaphorizontal": 0.1234,
            "slitgapvertical": 0.2345,
            "synchrotronmode": "User",
            "undulatorgap1": 1.234,
            "resolution": 1.1830593331191241,
            "wavelength": 1.11647184541378,
        },
    )
    mx.update_data_collection_append_comments.assert_called_with(
        TEST_DATA_COLLECTION_IDS[0], "Sample position (µm): (158, 24, 3)", " "
    )



from ophyd_async.core import YamlSettingsProvider
from ophyd_async.fastcs.panda import HDFPanda
from ophyd_async.plan_stubs import apply_panda_settings, retrieve_settings


def load_panda_from_yaml(yaml_directory: str, yaml_file_name: str, panda: HDFPanda):
    provider = YamlSettingsProvider(yaml_directory)
    settings = yield from retrieve_settings(provider, yaml_file_name, panda)
    yield from apply_panda_settings(settings)



from datetime import datetime
from unittest.mock import patch

import pydantic
import pytest
from bluesky.simulators import assert_message_and_return_remaining
from dodal.devices.aperturescatterguard import ApertureScatterguard
from dodal.devices.areadetector.plugins.CAM import ColorMode
from dodal.devices.backlight import Backlight
from dodal.devices.oav.oav_detector import OAV
from dodal.devices.oav.oav_parameters import OAVParameters
from dodal.devices.smargon import Smargon

from mx_bluesky.common.parameters.components import WithSnapshot
from mx_bluesky.common.parameters.constants import DocDescriptorNames
from mx_bluesky.hyperion.experiment_plans.oav_snapshot_plan import (
    OAV_SNAPSHOT_SETUP_SHOT,
    OavSnapshotComposite,
    oav_snapshot_plan,
)

from ....conftest import raw_params_from_file


@pytest.fixture
def oav_snapshot_params():
    return WithSnapshot(
        **raw_params_from_file(
            "tests/test_data/parameter_json_files/test_oav_snapshot_params.json"
        )
    )


@pydantic.dataclasses.dataclass(config={"arbitrary_types_allowed": True})
class CompositeImpl(OavSnapshotComposite):
    smargon: Smargon
    oav: OAV
    aperture_scatterguard: ApertureScatterguard
    backlight: Backlight


@pytest.fixture
def oav_snapshot_composite(smargon, oav, aperture_scatterguard, backlight):
    return CompositeImpl(
        smargon=smargon,
        oav=oav,
        aperture_scatterguard=aperture_scatterguard,
        backlight=backlight,
    )


@patch("mx_bluesky.hyperion.experiment_plans.oav_snapshot_plan.datetime", spec=datetime)
def test_oav_snapshot_plan_issues_rotations_and_generates_events(
    mock_datetime, oav_snapshot_params, oav_snapshot_composite, sim_run_engine
):
    mock_datetime.now.return_value = datetime.fromisoformat("2024-06-07T10:06:23.12")
    msgs = sim_run_engine.simulate_plan(
        oav_snapshot_plan(
            oav_snapshot_composite,
            oav_snapshot_params,
            OAVParameters(oav_config_json="tests/test_data/test_OAVCentring.json"),
        )
    )

    msgs = assert_message_and_return_remaining(
        msgs,
        lambda msg: msg.command == "set"
        and msg.obj.name == "oav-cam-color_mode"
        and msg.args[0] == ColorMode.RGB1,
    )
    msgs = assert_message_and_return_remaining(
        msgs,
        lambda msg: msg.command == "set"
        and msg.obj.name == "oav-cam-acquire_period"
        and msg.args[0] == 0.05,
    )
    msgs = assert_message_and_return_remaining(
        msgs,
        lambda msg: msg.command == "set"
        and msg.obj.name == "oav-cam-acquire_time"
        and msg.args[0] == 0.075,
    )
    msgs = assert_message_and_return_remaining(
        msgs,
        lambda msg: msg.command == "set"
        and msg.obj.name == "oav-cam-gain"
        and msg.args[0] == 1,
    )
    msgs = assert_message_and_return_remaining(
        msgs,
        lambda msg: msg.command == "set"
        and msg.obj.name == "oav-zoom_controller"
        and msg.args[0] == "5.0x",
    )
    msgs = assert_message_and_return_remaining(
        msgs,
        lambda msg: msg.command == "set"
        and msg.obj.name == "oav-snapshot-directory"
        and msg.args[0] == "/tmp/my_snapshots",
    )
    for expected in [
        {"omega": 0, "filename": "10062312_oav_snapshot_0"},
        {"omega": 90, "filename": "10062312_oav_snapshot_90"},
        {"omega": 180, "filename": "10062312_oav_snapshot_180"},
        {"omega": 270, "filename": "10062312_oav_snapshot_270"},
    ]:
        msgs = assert_message_and_return_remaining(
            msgs,
            lambda msg: msg.command == "set"
            and msg.obj.name == "smargon-omega"
            and msg.args[0] == expected["omega"]
            and msg.kwargs["group"] == OAV_SNAPSHOT_SETUP_SHOT,
        )
        msgs = assert_message_and_return_remaining(
            msgs,
            lambda msg: msg.command == "set"
            and msg.obj.name == "oav-snapshot-filename"
            and msg.args[0] == expected["filename"],
        )
        msgs = assert_message_and_return_remaining(
            msgs,
            lambda msg: msg.command == "trigger"
            and msg.obj.name == "oav-snapshot"
            and msg.kwargs["group"] is None,
        )
        msgs = assert_message_and_return_remaining(
            msgs,
            lambda msg: msg.command == "wait" and msg.kwargs["group"] is None,
        )
        msgs = assert_message_and_return_remaining(
            msgs,
            lambda msg: msg.command == "create"
            and msg.kwargs["name"]
            == DocDescriptorNames.OAV_ROTATION_SNAPSHOT_TRIGGERED,
        )
        msgs = assert_message_and_return_remaining(
            msgs, lambda msg: msg.command == "read" and msg.obj.name == "oav"
        )
        msgs = assert_message_and_return_remaining(
            msgs, lambda msg: msg.command == "save"
        )



from unittest.mock import MagicMock, call

import pytest
from bluesky.run_engine import RunEngine
from bluesky.simulators import RunEngineSimulator, assert_message_and_return_remaining
from dodal.devices.focusing_mirror import (
    FocusingMirrorWithStripes,
    MirrorStripe,
    MirrorVoltages,
)
from dodal.devices.undulator_dcm import UndulatorDCM
from ophyd_async.testing import get_mock_put

from mx_bluesky.hyperion.device_setup_plans import dcm_pitch_roll_mirror_adjuster
from mx_bluesky.hyperion.device_setup_plans.dcm_pitch_roll_mirror_adjuster import (
    YAW_LAT_TIMEOUT_S,
    adjust_dcm_pitch_roll_vfm_from_lut,
    adjust_mirror_stripe,
)


def test_when_bare_mirror_stripe_selected_then_expected_voltages_set_and_waited(
    sim_run_engine: RunEngineSimulator,
    mirror_voltages: MirrorVoltages,
):
    messages = sim_run_engine.simulate_plan(
        dcm_pitch_roll_mirror_adjuster._apply_and_wait_for_voltages_to_settle(
            MirrorStripe.BARE, mirror_voltages
        )
    )

    for channel, expected_voltage in zip(
        mirror_voltages.horizontal_voltages.values(),
        [1, 107, 15, 139, 41, 165, 11, 6, 166, -65, 0, -38, 179, 128],
        strict=True,
    ):
        messages = assert_message_and_return_remaining(
            messages,
            lambda msg: msg.command == "set"
            and msg.obj == channel
            and msg.args[0] == expected_voltage,
        )
        messages = assert_message_and_return_remaining(
            messages, lambda msg: msg.command == "wait"
        )

    for channel, expected_voltage in zip(
        mirror_voltages.vertical_voltages.values(),
        [140, 100, 70, 30, 30, -65, 24, 15],
        strict=True,
    ):
        messages = assert_message_and_return_remaining(
            messages,
            lambda msg: msg.command == "set"
            and msg.obj == channel
            and msg.args[0] == expected_voltage,
        )
        messages = assert_message_and_return_remaining(
            messages, lambda msg: msg.command == "wait"
        )


@pytest.mark.parametrize(
    "energy_kev, expected_stripe, expected_lat, expected_yaw, first_voltage, last_voltage",
    [
        (6.999, MirrorStripe.BARE, 0.0, 6.2, 140, 15),
        (7.001, MirrorStripe.RHODIUM, 10.0, 0.0, 124, -46),
    ],
)
def test_adjust_mirror_stripe(
    RE: RunEngine,
    mirror_voltages: MirrorVoltages,
    vfm: FocusingMirrorWithStripes,
    energy_kev,
    expected_stripe,
    expected_lat,
    expected_yaw,
    first_voltage,
    last_voltage,
):
    parent = MagicMock()
    parent.attach_mock(get_mock_put(vfm.stripe), "stripe_set")
    parent.attach_mock(get_mock_put(vfm.apply_stripe), "apply_stripe")
    parent.attach_mock(get_mock_put(vfm.x_mm.user_setpoint), "lat_set")
    parent.attach_mock(get_mock_put(vfm.yaw_mrad.user_setpoint), "yaw_mrad")

    RE(adjust_mirror_stripe(energy_kev, vfm, mirror_voltages))

    expected_calls = [
        call.stripe_set(expected_stripe, wait=True),
        call.apply_stripe(None, wait=True),
        call.lat_set(expected_lat, wait=True),
        call.yaw_mrad(expected_yaw, wait=True),
    ]
    assert parent.method_calls == expected_calls
    mirror_voltages.vertical_voltages[0].set.assert_called_once_with(  # type: ignore
        first_voltage
    )
    mirror_voltages.vertical_voltages[7].set.assert_called_once_with(  # type: ignore
        last_voltage
    )


def test_adjust_dcm_pitch_roll_vfm_from_lut(
    undulator_dcm: UndulatorDCM,
    vfm: FocusingMirrorWithStripes,
    mirror_voltages: MirrorVoltages,
    sim_run_engine: RunEngineSimulator,
):
    sim_run_engine.add_handler_for_callback_subscribes()

    messages = sim_run_engine.simulate_plan(
        adjust_dcm_pitch_roll_vfm_from_lut(undulator_dcm, vfm, mirror_voltages, 7.5)
    )
    # target bragg angle 15.288352 deg
    messages = assert_message_and_return_remaining(
        messages,
        lambda msg: msg.command == "set"
        and msg.obj.name == "dcm-pitch_in_mrad"
        and abs(msg.args[0] - -0.78229639) < 1e-5
        and msg.kwargs["group"] == "DCM_GROUP",
    )
    messages = assert_message_and_return_remaining(
        messages[1:],
        lambda msg: msg.command == "set"
        and msg.obj.name == "dcm-roll_in_mrad"
        and abs(msg.args[0] - -0.2799) < 1e-5
        and msg.kwargs["group"] == "DCM_GROUP",
    )
    messages = assert_message_and_return_remaining(
        messages[1:],
        lambda msg: msg.command == "set"
        and msg.obj.name == "vfm-stripe"
        and msg.args == (MirrorStripe.RHODIUM,),
    )
    messages = assert_message_and_return_remaining(
        messages[1:],
        lambda msg: msg.command == "wait",
    )
    messages = assert_message_and_return_remaining(
        messages[1:],
        lambda msg: msg.command == "trigger" and msg.obj.name == "vfm-apply_stripe",
    )
    messages = assert_message_and_return_remaining(
        messages[1:],
        lambda msg: msg.command == "set"
        and msg.obj is vfm.x_mm
        and msg.args == (10.0,)
        and msg.kwargs["timeout"] == YAW_LAT_TIMEOUT_S,
    )
    messages = assert_message_and_return_remaining(
        messages[1:], lambda msg: msg.command == "wait"
    )
    messages = assert_message_and_return_remaining(
        messages[1:],
        lambda msg: msg.command == "set"
        and msg.obj is vfm.yaw_mrad
        and msg.args == (0.0,)
        and msg.kwargs["timeout"] == YAW_LAT_TIMEOUT_S,
    )
    messages = assert_message_and_return_remaining(
        messages[1:], lambda msg: msg.command == "wait"
    )
    for channel, expected_voltage in enumerate(
        [11, 117, 25, 149, 51, 145, -9, -14, 146, -10, 55, 17, 144, 93]
    ):
        messages = assert_message_and_return_remaining(
            messages[1:],
            lambda msg: msg.command == "set"
            and msg.obj.name == f"mirror_voltages-horizontal_voltages-{channel}"
            and msg.args == (expected_voltage,),
        )
    for channel, expected_voltage in enumerate([124, 114, 34, 49, 19, -116, 4, -46]):
        messages = assert_message_and_return_remaining(
            messages[1:],
            lambda msg: msg.command == "set"
            and msg.obj.name == f"mirror_voltages-vertical_voltages-{channel}"
            and msg.args == (expected_voltage,),
        )
